@@ -33,7 +33,7 @@
 #include "cardmgr.h"
 
 static const char *version =
-    "cardmgr.c 1.53 1995/04/11 18:26:40 (David Hinds)";
+    "cardmgr.c 1.55 1995/05/15 01:48:24 (David Hinds)";
 
 /*====================================================================*/
 
@@ -42,6 +42,7 @@ typedef struct socket_info_t {
     int state;
     card_info_t *card;
     bind_info_t bind[MAX_FUNCTIONS];
+    char *mtd[2*CISTPL_MAX_DEVICES];
 } socket_info_t;
 
 #define SOCKET_PRESENT	0x01
@@ -63,6 +64,12 @@ struct card_info_t *root_card = NULL;
 /* Linked list of function definitions */
 struct card_info_t *root_func = NULL;
 
+/* Linked list of MTD definitions */
+struct mtd_ident_t *root_mtd = NULL;
+
+/* Default MTD */
+struct mtd_ident_t *default_mtd = NULL;
+
 static int sockets;
 static struct socket_info_t socket[MAX_SOCKS];
 
@@ -345,19 +352,15 @@ static void free_card(card_info_t *card)
 
 static void free_config(void)
 {
-    adjust_list_t *adj;
-    device_info_t *dev;
-    card_info_t *card;
-    int i;
-    
     while (root_adjust != NULL) {
-	adj = root_adjust;
+	adjust_list_t *adj = root_adjust;
 	root_adjust = root_adjust->next;
 	free(adj);
     }
     
     while (root_device != NULL) {
-	dev = root_device;
+	device_info_t *dev = root_device;
+	int i;
 	root_device = root_device->next;
 	for (i = 0; i < dev->modules; i++) {
 	    free(dev->module[i]);
@@ -369,17 +372,26 @@ static void free_config(void)
     }
 
     while (root_card != NULL) {
-	card = root_card;
+	card_info_t *card = root_card;
 	root_card = root_card->next;
 	free_card(card);
     }
+    
     while (root_func != NULL) {
-	card = root_func;
+	card_info_t *card = root_func;
 	root_func = root_func->next;
 	free_card(card);
     }
     blank_card = NULL;
     
+    while (root_mtd != NULL) {
+	mtd_ident_t *mtd = root_mtd;
+	root_mtd = root_mtd->next;
+	free(mtd->name);
+	free(mtd->module);
+	free(mtd);
+    }
+    default_mtd = NULL;
 } /* free_config */
 
 /*====================================================================*/
@@ -507,6 +519,88 @@ static void remove_module(char *mod)
 
 /*====================================================================*/
 
+static mtd_ident_t *lookup_mtd(region_info_t *region)
+{
+    mtd_ident_t *mtd;
+    int match = 0;
+    
+    for (mtd = root_mtd; mtd; mtd = mtd->next) {
+	switch (mtd->mtd_type) {
+	case JEDEC_MTD:
+	    if ((mtd->jedec_mfr == region->JedecMfr) &&
+		(mtd->jedec_info == region->JedecInfo)) {
+		match = 1;
+		break;
+	    }
+	case DTYPE_MTD:
+	    break;
+	default:
+	    break;
+	}
+	if (match) break;
+    }
+    if (mtd)
+	return mtd;
+    else
+	return default_mtd;
+} /* lookup_mtd */
+
+/*====================================================================*/
+
+static void bind_mtd(int sn)
+{
+    socket_info_t *s = &socket[sn];
+    region_info_t region;
+    bind_info_t bind;
+    mtd_info_t mtd_info;
+    mtd_ident_t *mtd;
+    int i, attr, ret, nr;
+
+    nr = 0;
+    for (attr = 0; attr < 2; attr++) {
+	region.Attributes = attr;
+	ret = ioctl(s->fd, DS_GET_FIRST_REGION, &region);
+	while (ret == 0) {
+	    mtd = lookup_mtd(&region);
+	    if (mtd) {
+		/* Have we seen this MTD before? */
+		for (i = 0; i < nr; i++)
+		    if (strcmp(s->mtd[i], mtd->module) == 0) break;
+		if (i == nr) {
+		    install_module(mtd->module, NULL);
+		    s->mtd[nr] = mtd->module;
+		    nr++;
+		}
+		syslog(LOG_INFO, "  %s memory region at 0x%lx: %s",
+		       attr ? "Attribute" : "Common", region.CardOffset,
+		       mtd->name);
+		/* Bind MTD to this region */
+		strcpy(mtd_info.dev_info, s->mtd[i]);
+		mtd_info.Attributes = region.Attributes;
+		mtd_info.CardOffset = region.CardOffset;
+		if (ioctl(s->fd, DS_BIND_MTD, &mtd_info) != 0) {
+		    syslog(LOG_INFO,
+			   "bind MTD '%s' to region at 0x%lx failed: %m",
+			   (char *)mtd_info.dev_info, region.CardOffset);
+		}
+	    }
+	    ret = ioctl(s->fd, DS_GET_NEXT_REGION, &region);
+	}
+    }
+    s->mtd[nr] = NULL;
+    
+    /* Now bind each unique MTD as a normal client of this socket */
+    for (i = 0; i < nr; i++) {
+	strcpy(bind.dev_info, s->mtd[i]);
+	bind.instance = NULL;
+	if (ioctl(s->fd, DS_BIND_REQUEST, &bind) != 0)
+	    syslog(LOG_INFO, "bind MTD '%s' to socket %d failed: %m",
+		   (char *)bind.dev_info, sn);
+    }
+} /* bind_mtd */
+
+/*====================================================================*/
+
 static void do_insert(int sn)
 {
     socket_info_t *s = &socket[sn];
@@ -534,13 +628,16 @@ static void do_insert(int sn)
 	bind[i].name[0] = '\0';
     }
     write_stab();
+
+    /* Set up MTD's */
+    bind_mtd(sn);
     
     /* Install kernel modules */
     for (i = 0; i < card->functions; i++) {
 	for (j = 0; j < dev[i]->modules; j++)
 	    install_module(dev[i]->module[j], dev[i]->opts[j]);
     }
-    
+
     /* Bind drivers by their dev_info identifiers */
     for (i = 0; i < card->functions; i++) {
 	if (ioctl(s->fd, DS_BIND_REQUEST, &bind[i]) != 0) {
@@ -626,6 +723,9 @@ static int do_remove(int sn, int request)
 	for (j = dev[i]->modules-1; j >= 0; j--)
 	    remove_module(dev[i]->module[j]);
     }
+    /* Remove any MTD's bound to this socket */
+    for (i = 0; (s->mtd[i] != NULL); i++)
+	remove_module(s->mtd[i]);
 
 done:
     beep(BEEP_TIME, BEEP_OK);
