@@ -96,7 +96,7 @@ static char *if_names[] = {"10baseT", "AUI", "undefined", "BNC"};
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-    "3c589_cs.c 1.41 1995/04/11 18:25:31 (David Hinds)\n";
+    "3c589_cs.c 1.43 1995/05/15 04:46:01 (David Hinds)\n";
 #endif
 
 #ifdef TC589_DEBUG
@@ -336,6 +336,7 @@ static void tc589_config(dev_link_t *link)
 #endif
 
     do {
+	tuple.Attributes = 0;
 	tuple.DesiredTuple = CISTPL_CONFIG;
 	i = CardServices(GetFirstTuple, handle, &tuple);
 	if (i != CS_SUCCESS) break;
@@ -974,7 +975,7 @@ int init_module(void)
 #endif
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("serial_cs: Card Services release does not match!");
+	printk("serial_cs: Card Services release does not match!\n");
 	return -1;
     }
     register_pcmcia_driver(&dev_info, &tc589_attach, &tc589_detach);
