@@ -46,6 +46,8 @@ static int do_mtd_request(memory_handle_t handle, mtd_request_t *req,
     socket_info_t *s;
     
     mtd = handle->mtd;
+    if (mtd == NULL)
+	return CS_GENERAL_FAILURE;
     s = SOCKET(mtd);
     for (;;) {
 	mtd->event_callback_args.mtdrequest = req;
@@ -72,6 +74,7 @@ static int do_mtd_request(memory_handle_t handle, mtd_request_t *req,
 	    break;
 	}
 	if (current->signal & ~current->blocked) {
+	    printk("cs: do_mtd_request interrupted!\n");
 	    ret = CS_GENERAL_FAILURE;
 	    break;
 	}
@@ -279,6 +282,21 @@ static int mtd_set_vpp(client_handle_t handle, mtd_vpp_req_t *req)
     return CS_SUCCESS;
 }
 
+static int mtd_rdy_mask(client_handle_t handle, mtd_rdy_req_t *req)
+{
+    socket_info_t *s;
+    if (CHECK_HANDLE(handle))
+	return CS_BAD_HANDLE;
+    s = SOCKET(handle);
+    if (req->Mask & CS_EVENT_READY_CHANGE)
+	s->socket.csc_mask |= SS_READY;
+    else
+	s->socket.csc_mask &= ~SS_READY;
+    if (s->ss_entry(s->sock, SS_SetSocket, &s->socket))
+	return CS_GENERAL_FAILURE;
+    return CS_SUCCESS;
+}
+
 int MTDHelperEntry(int func, void *a1, void *a2)
 {
     switch (func) {
@@ -286,6 +304,8 @@ int MTDHelperEntry(int func, void *a1, void *a2)
 	return mtd_modify_window(a1, a2); break;
     case MTDSetVpp:
 	return mtd_set_vpp(a1, a2); break;
+    case MTDRDYMask:
+	return mtd_rdy_mask(a1, a2); break;
     default:
 	return CS_UNSUPPORTED_FUNCTION; break;
     }
@@ -339,13 +359,13 @@ static void setup_regions(client_handle_t handle, int attr,
     code = (attr) ? CISTPL_JEDEC_A : CISTPL_JEDEC_C;
     has_jedec = (read_tuple(handle, code, &jedec) == CS_SUCCESS);
     if (has_jedec && (device.ndev != jedec.nid)) {
-	printk("cs: Device info does not match JEDEC info!\n");
+	printk("cs: Device info does not match JEDEC info.\n");
 	has_jedec = 0;
     }
     code = (attr) ? CISTPL_DEVICE_GEO_A : CISTPL_DEVICE_GEO;
     has_geo = (read_tuple(handle, code, &geo) == CS_SUCCESS);
     if (has_geo && (device.ndev != geo.ngeo)) {
-	printk("cs: Device info does not match geometry tuple!\n");
+	printk("cs: Device info does not match geometry tuple.\n");
 	has_geo = 0;
     }
     
@@ -562,8 +582,8 @@ int open_memory(client_handle_t *handle, open_mem_t *open)
 int close_memory(memory_handle_t handle)
 {
 #ifdef PCMCIA_DEBUG
-	if (pc_debug)
-	    printk ("cs: close_memory(0x%p)\n", handle);
+    if (pc_debug)
+	printk ("cs: close_memory(0x%p)\n", handle);
 #endif
     if (CHECK_REGION(handle))
 	return CS_BAD_HANDLE;
