@@ -54,6 +54,10 @@
 #define DE650_START_PG	0x40	/* First page of TX buffer */
 #define DE650_STOP_PG	0x80	/* Last page +1 of RX ring */
 
+/* Socket EA cards have a larger packet buffer */
+#define SOCKET_START_PG	0x01
+#define SOCKET_STOP_PG	0xff
+
 #define DE650_RDC_TIMEOUT 0x02	/* Max wait in jiffies for Tx RDC */
 
 static char *if_names[] = {"10baseT", "undefined", "undefined", "BNC"};
@@ -61,7 +65,7 @@ static char *if_names[] = {"10baseT", "undefined", "undefined", "BNC"};
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-    "de650_cs.c 1.65 1995/05/15 04:46:01 (David Hinds)\n";
+    "de650_cs.c 1.67 1995/05/24 15:38:36 (David Hinds)\n";
 #endif
 
 /*====================================================================*/
@@ -74,6 +78,9 @@ static u_long irq_mask = 0xdeb8;
 /* Transceiver type, for Socket EA cards */
 static int if_port = 3;
 
+/* Use 64K packet buffer, for Socket EA cards */
+static int do_big_buf = 1;
+
 /* Insert a pause in block_output, for Socket EA and Accton cards */
 static int delay_output = 0;
 
@@ -124,7 +131,8 @@ static hwaddr_info_t hw_info[] = {
     { 0x0ff0, 0x00, 0x00, 0xe8, "Accton EN2212" },
     { 0x4000, 0x00, 0xc0, 0x1b, "Socket EA" },
     { 0x5000, 0x00, 0x00, 0xe8, "Maxtech PCN2000" },
-    { 0x0110, 0x00, 0x40, 0xf6, "Katron PE-520" }
+    { 0x0110, 0x00, 0x40, 0xf6, "Katron PE-520" },
+    { 0x0ff0, 0x00, 0xa0, 0x0c, "NE2000 Compatible" }
 };
 
 #define IS_ACCTON(dev) (ei_status.name == hw_info[4].id)
@@ -442,6 +450,13 @@ static void de650_config(dev_link_t *link)
 
     if (IS_SOCKET(dev) || IS_ACCTON(dev))
 	delay_output = 1;
+
+    /* Use 64K packet buffer for Socket EA cards */
+    if (IS_SOCKET(dev) && do_big_buf) {
+	ei_status.tx_start_page = SOCKET_START_PG;
+	ei_status.rx_start_page = ei_status.tx_start_page + TX_PAGES;
+	ei_status.stop_page = SOCKET_STOP_PG;
+    }
     
     printk("%s: %s, port %#3lx, irq %d,",
 	   dev->name, id, dev->base_addr, dev->irq);
@@ -589,6 +604,8 @@ static int de650_open(struct device *dev)
 	tmp = inb_p(DE650_BASE+SOCKET_IFPORT) & ~3;
 	if (dev->if_port == 3)
 	    tmp |= 1;
+	if (do_big_buf)
+	    tmp |= 2;
 	outb_p(tmp, DE650_BASE+SOCKET_IFPORT);
     }
     
@@ -650,6 +667,8 @@ static void de650_reset_8390(struct device *dev)
 	tmp = inb_p(DE650_BASE+SOCKET_IFPORT) & ~3;
 	if (dev->if_port == 3)
 	    tmp |= 1;
+	if (do_big_buf)
+	    tmp |= 2;
 	outb_p(tmp, DE650_BASE+SOCKET_IFPORT);
     }
 	    
