@@ -42,7 +42,7 @@
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static const char *version =
-    "i82365.c 1.65 1995/04/11 18:25:08 (David Hinds)\n";
+    "i82365.c 1.68 1995/05/15 01:44:37 (David Hinds)\n";
 #endif
 
 /*====================================================================*/
@@ -79,8 +79,8 @@ static int fast_pci = 0;
 
 /* Timing parameters for Cirrus PD67XX ISA chips */
 static int freq_bypass = 0;
-static int setup_time = 0;
-static int cmd_time = 0;
+static int setup_time = 1;
+static int cmd_time = 6;
 static int recov_time = 0;
 static int wakeup = 0;
 
@@ -273,6 +273,10 @@ static int identify(int sock)
 	type = IS_VLSI; break;
     case 0x88: case 0x89:
 	type = IS_IBM | (val & 7); break;
+    case 0x8b:
+	type = IS_VG468 | (val & 7);
+	i365_bclr(sock, VG468_MISC, VG468_MISC_VADEMREV);
+	break;
     default:
 	return -1;
     }
@@ -823,6 +827,8 @@ static int i365_set_socket(u_short lsock, socket_state_t *state)
     case IS_VG468:
 	if (async_clock)
 	    i365_bset(psock, VG468_CTL, VG468_CTL_ASYNC);
+	else
+	    i365_bclr(psock, VG468_CTL, VG468_CTL_ASYNC);
 	break;
 #ifdef CONFIG_PCI
     case IS_PD6729:
@@ -859,18 +865,14 @@ static int i365_set_socket(u_short lsock, socket_state_t *state)
 	/* Timing parameters */
 	if (freq_bypass)
 	    i365_bset(psock, PD67_MISC_CTL_2, PD67_MC2_FREQ_BYPASS);
-	if (setup_time) {
-	    i365_set(psock, PD67_TIME_SETUP(0), setup_time);
-	    i365_set(psock, PD67_TIME_SETUP(1), setup_time);
-	}
-	if (cmd_time) {
-	    i365_set(psock, PD67_TIME_CMD(0), cmd_time);
-	    i365_set(psock, PD67_TIME_CMD(1), cmd_time + 9);
-	}
-	if (recov_time) {
-	    i365_set(psock, PD67_TIME_RECOV(0), recov_time);
-	    i365_set(psock, PD67_TIME_RECOV(1), recov_time);
-	}
+	else
+	    i365_bclr(psock, PD67_MISC_CTL_2, PD67_MC2_FREQ_BYPASS);
+	i365_set(psock, PD67_TIME_SETUP(0), setup_time);
+	i365_set(psock, PD67_TIME_SETUP(1), setup_time);
+	i365_set(psock, PD67_TIME_CMD(0), cmd_time);
+	i365_set(psock, PD67_TIME_CMD(1), cmd_time + 9);
+	i365_set(psock, PD67_TIME_RECOV(0), recov_time);
+	i365_set(psock, PD67_TIME_RECOV(1), recov_time);
 	break;
     default:
 	break;
@@ -1086,7 +1088,7 @@ int init_module(void)
     servinfo_t serv;
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("i82365: Card Services release does not match!");
+	printk("i82365: Card Services release does not match!\n");
 	return -1;
     }
     return i365_init();
