@@ -51,7 +51,7 @@
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-    "ibmcc_cs.c 1.31 1995/04/11 18:25:31 (David Hinds)\n";
+    "ibmcc_cs.c 1.33 1995/05/15 04:46:01 (David Hinds)\n";
 #endif
 
 static char *if_names[] = {"10baseT", "undefined", "undefined", "BNC"};
@@ -280,6 +280,7 @@ static void ibmcc_config(dev_link_t *link)
 #endif
 
     do {
+	tuple.Attributes = 0;
 	tuple.DesiredTuple = CISTPL_CONFIG;
 	i = CardServices(GetFirstTuple, handle, &tuple);
 	if (i != CS_SUCCESS) break;
@@ -693,7 +694,7 @@ int init_module(void)
 #endif
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("ibmcc_cs: Card Services release does not match!");
+	printk("ibmcc_cs: Card Services release does not match!\n");
 	return -1;
     }
     register_pcmcia_driver(&dev_info, &ibmcc_attach, &ibmcc_detach);
