@@ -13,7 +13,7 @@
 
 #include <linux/config.h>
 
-#define PCMCIA_DEBUG 1
+/* #define PCMCIA_DEBUG 1 */
 
 #ifdef MODULE
 #include <linux/module.h>
@@ -26,6 +26,7 @@
 #include <linux/malloc.h>
 #include <linux/string.h>
 #include <linux/timer.h>
+#include <linux/timex.h>
 #include <linux/major.h>
 #include <linux/fs.h>
 #include <asm/io.h>
@@ -44,7 +45,7 @@
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-"iflash2+_mtd.c 1.5 1995/05/15 01:42:03 (David Hinds)\n";
+"iflash2+_mtd.c 1.8 1995/05/29 22:34:53 (David Hinds)\n";
 #endif
 
 static int vpp_timeout_period	= 1000;	/* in ticks */
@@ -52,9 +53,9 @@ static int vpp_settle		= 10;	/* in ticks */
 static int erase_timeout	= 100;	/* in ms */
 static int erase_limit		= 1000;	/* in ticks */
 
-static int write_limit		= 4;	/* max retries */
+static int retry_limit		= 4;	/* max retries */
 
-static u_long max_tries       	= 4096*256;
+static u_long max_tries       	= 4096;
 
 /*====================================================================*/
 
@@ -80,6 +81,7 @@ typedef struct flash_region_t {
 	u_long		state;
 	u_long		erase_time;
 	u_long		erase_addr;
+	u_long		erase_retries;
     } cell[MAX_CELLS];
 } flash_region_t;
 
@@ -90,6 +92,7 @@ typedef struct flash_dev_t {
     int			vpp_usage;
     u_long		vpp_start;
     struct timer_list	vpp_timeout;
+    struct wait_queue	write_queue;
     flash_region_t	*flash[2*CISTPL_MAX_DEVICES];
 } flash_dev_t;
 
@@ -101,6 +104,38 @@ static dev_info_t dev_info = "iflash2+_mtd";
 
 static dev_link_t *dev_list = NULL;
 
+/*====================================================================*/
+
+static void kmsg(char *fmt, ...)
+{
+    char buf[256];
+    va_list args;
+    
+    va_start(args, fmt);
+    vsprintf(buf, fmt, args);
+    va_end(args);
+    printk("%s: %s\n", (char *)dev_info, buf);
+}
+
+static void cs_error(int func, int ret)
+{
+    CardServices(ReportError, dev_info, (void *)func, (void *)ret);
+}
+
+#ifdef PCMCIA_DEBUG
+static inline u_long uticks(void)
+{
+    u_long count;
+    outb_p(0x00, 0x43);
+    count = inb_p(0x40);
+    count |= inb(0x40) << 8;
+    count = ((LATCH-1) - count) * 10000;
+    count = (count + LATCH/2) / LATCH;
+    count += jiffies * 10000;
+    return count;
+}
+#endif
+
 /*======================================================================
 
     Low level routines for programming the flash card.
@@ -123,23 +158,65 @@ static void log_esr(char *s, volatile u_short *esr)
 	printk("CSR = 0x%04x, BSR = 0x%04x, GSR = 0x%04x\n",
 	       CSR, BSR, GSR);
 }
-    
-static int check_write(volatile u_short *esr)
+
+static void abort_cmd(volatile u_short *esr)
 {
-    u_short BSR = 0;
-    u_long i;
+    u_short i;
+    
+    *esr = IF_ABORT;
+    *esr = IF_READ_ESR;
+    for (i = 0; i < max_tries; i++)
+	if ((esr[4] & GSR_SLEEP) == GSR_SLEEP) break;
+    if (i == max_tries)
+	kmsg("abort cmd failed!");
+    *esr = IF_READ_ARRAY;
+    *esr = IF_CLEAR_CSR;
+}
+
+static int set_rdy_mode(volatile u_short *esr, u_short mode)
+{
+    u_short i;
     
     *esr = IF_READ_ESR;
-    for (i = 0; i < max_tries; i++) {
-	BSR = esr[2];
-	if ((BSR & BSR_READY) == BSR_READY) break;
-    }
+    for (i = 0; i < max_tries; i++)
+	if (!(esr[2] & BSR_QUEUE_FULL)) break;
     if (i == max_tries) {
-	printk("check_write: timed out!\n");
+	kmsg("set_rdy_mode timed out!");
 	log_esr(NULL, esr);
+	abort_cmd(esr);
 	return CS_GENERAL_FAILURE;
     }
-    if (BSR & BSR_FAILED) {
+    *esr = IF_RDY_MODE;
+    *esr = mode;
+    *esr = IF_READ_ESR;
+    for (i = 0; i < max_tries; i++) {
+	if ((esr[4] & GSR_WR_READY) == GSR_WR_READY) break;
+    }
+    if (i == max_tries) {
+	kmsg("set_rdy_mode timed out!");
+	log_esr(NULL, esr);
+	abort_cmd(esr);
+	return CS_GENERAL_FAILURE;
+    }
+    if (esr[4] & GSR_OP_ERR)
+	return CS_GENERAL_FAILURE;
+    return CS_SUCCESS;
+}
+
+static int check_write(struct wait_queue **queue, volatile u_short *esr)
+{
+    *esr = IF_READ_ESR;
+    if ((esr[2] & BSR_READY) != BSR_READY) {
+	current->timeout = jiffies + 2;
+	interruptible_sleep_on(queue);
+	if ((esr[2] & BSR_READY) != BSR_READY) {
+	    kmsg("check_write: timed out!");
+	    log_esr(NULL, esr);
+	    abort_cmd(esr);
+	    return CS_GENERAL_FAILURE;
+	}
+    }
+    if (esr[2] & BSR_FAILED) {
 	log_esr("write error", esr);
 	return CS_WRITE_FAILURE;
     }
@@ -147,34 +224,38 @@ static int check_write(volatile u_short *esr)
 	return CS_SUCCESS;
 }
 
-static int page_setup(volatile u_short *esr, volatile u_short *address,
-		      u_short count)
+static int page_setup(struct wait_queue **queue, volatile u_short *esr,
+		      volatile u_short *address,  u_short count)
 {
-    u_short GSR, nw;
-    u_long i;
+    u_short nw;
     
     *esr = IF_READ_ESR;
-    for (i = 0; i < max_tries; i++) {
-	GSR = esr[4];
-	if ((GSR & GSR_PAGE_AVAIL) == GSR_PAGE_AVAIL) break;
-    }
-    if (i == max_tries) {
-	printk("page_setup timed out\n");
-	log_esr(NULL, esr);
-	return CS_GENERAL_FAILURE;
+    if ((esr[4] & GSR_PAGE_AVAIL) != GSR_PAGE_AVAIL) {
+	current->timeout = jiffies + 2;
+	interruptible_sleep_on(queue);
+	if ((esr[4] & GSR_PAGE_AVAIL) != GSR_PAGE_AVAIL) {
+	    kmsg("page_setup timed out");
+	    log_esr(NULL, esr);
+	    return CS_GENERAL_FAILURE;
+	}
     }
     nw = count >> 1;
     if (nw == 0) {
 	/* Special case of single byte write */
-	*(u_char *)address = LOW(IF_SEQ_LOAD);
-	*(u_char *)address = 1;
-	*(u_char *)address = 0;
+	*(volatile u_char *)address = LOW(IF_SINGLE_LOAD);
     }
     else {
 	*address = IF_SEQ_LOAD;
 	/* Tricky: split count into low bytes and high bytes */
-	*address = (count-nw-1) | ((nw-1) << 8);
-	*address = 0;
+	nw = (count-nw-1) | ((nw-1) << 8);
+	if ((u_long)address & 2) {
+	    *address = 0;
+	    *address = nw;
+	}
+	else {
+	    *address = nw;
+	    *address = 0;
+	}
     }
     return CS_SUCCESS;
 }
@@ -182,29 +263,36 @@ static int page_setup(volatile u_short *esr, volatile u_short *address,
 static int page_write(volatile u_short *esr, volatile u_short *address,
 		      u_short count)
 {
-    u_short BSR;
-    u_long i;
+    u_short nw;
+    u_short i;
 
-    count = (count >> 1) - 1;
     *esr = IF_READ_ESR;
-    for (i = 0; i < max_tries; i++) {
-	BSR = esr[2];
-	if (!(BSR & BSR_QUEUE_FULL)) break;
-    }
+    for (i = 0; i < max_tries; i++)
+	if (!(esr[2] & BSR_QUEUE_FULL)) break;
     if (i == max_tries) {
-	printk("page_write timed out\n");
+	kmsg("page_write timed out");
 	log_esr(NULL, esr);
+	abort_cmd(esr);
 	return CS_GENERAL_FAILURE;
     }
-    if (count == 1) {
-	*(u_char *)address = LOW(IF_PAGE_WRITE);
-	*(u_char *)address = 1;
-	*(u_char *)address = 0;
+    nw = count >> 1;
+    if (nw == 0) {
+	*(volatile u_char *)address = LOW(IF_PAGE_WRITE);
+	*(volatile u_char *)address = 0;
+	*(volatile u_char *)address = 0;
     }
     else {
 	*address = IF_PAGE_WRITE;
-	*address = count | (count << 8);
-	*address = 0;
+	/* Tricky: split count into low bytes and high bytes */
+	nw = (count-nw-1) | ((nw-1) << 8);
+	if ((u_long)address & 2) {
+	    *address = 0;
+	    *address = nw;
+	}
+	else {
+	    *address = nw;
+	    *address = 0;
+	}
     }
     return CS_SUCCESS;
 }
@@ -232,17 +320,18 @@ static int check_erase(volatile u_short *address)
 
 static int suspend_erase(volatile u_short *esr)
 {
-    u_short GSR;
-    u_long i;
-    
+    u_short i;
+
+#ifdef PCMCIA_DEBUG
+    if (pc_debug > 1)
+	kmsg("suspending erase...");
+#endif
     *esr = IF_ERASE_SUSPEND;
     *esr = IF_READ_ESR;
-    for (i = 0; i < max_tries; i++) {
-	GSR = esr[4];
-	if ((GSR & GSR_WR_READY) == GSR_WR_READY) break;
-    }
+    for (i = 0; i < max_tries; i++)
+	if ((esr[4] & GSR_WR_READY) == GSR_WR_READY) break;
     if (i == max_tries) {
-	printk("suspend_erase timed out\n");
+	kmsg("suspend_erase timed out");
 	log_esr(NULL, esr);
 	return CS_GENERAL_FAILURE;
     }
@@ -252,18 +341,20 @@ static int suspend_erase(volatile u_short *esr)
 
 static void resume_erase(volatile u_short *esr)
 {
-    u_short GSR;
+#ifdef PCMCIA_DEBUG
+    if (pc_debug > 1)
+	kmsg("resuming erase...");
+#endif
     *esr = IF_READ_ESR;
-    GSR = esr[4];
     /* Only give resume signal if the erase is really suspended */
-    if (GSR & GSR_OP_SUSPEND)
+    if (esr[4] & GSR_OP_SUSPEND)
 	*esr = IF_CONFIRM;
 }
 
-static void reset_block(volatile u_short *address)
+static void reset_block(volatile u_short *esr)
 {
-    *address = IF_CLEAR_CSR;
-    *address = IF_READ_ARRAY;
+    *esr = IF_READ_ARRAY;
+    *esr = IF_CLEAR_CSR;
 }
 
 /*====================================================================*/
@@ -306,14 +397,17 @@ static int vpp_setup(dev_link_t *link, mtd_request_t *req)
 	dev->vpp_usage++;
 	/* If no other users, kill shutdown timer and apply power */
 	if (dev->vpp_usage == 1) {
+	    if (dev->vpp_timeout.expires)
+		del_timer(&dev->vpp_timeout);
+	    else {
 #ifdef PCMCIA_DEBUG
-	    if (pc_debug > 1)
-		printk("Raising Vpp...\n");
+		if (pc_debug > 1)
+		    kmsg("raising Vpp...");
 #endif
-	    del_timer(&dev->vpp_timeout);
-	    dev->vpp_start = jiffies;
-	    vpp_req.Vpp1 = vpp_req.Vpp2 = 120;
-	    MTDHelperEntry(MTDSetVpp, link->handle, &vpp_req);
+		dev->vpp_start = jiffies;
+		vpp_req.Vpp1 = vpp_req.Vpp2 = 120;
+		MTDHelperEntry(MTDSetVpp, link->handle, &vpp_req);
+	    }
 	}
     }
     /* Wait for Vpp to settle if it was just applied */
@@ -328,12 +422,15 @@ static int vpp_setup(dev_link_t *link, mtd_request_t *req)
 static void vpp_off(u_long arg)
 {
     dev_link_t *link = (dev_link_t *)arg;
+    flash_dev_t *dev;
     mtd_vpp_req_t req;
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("Lowering Vpp...\n");
+	kmsg("lowering Vpp...");
 #endif
+    dev = (flash_dev_t *)link->priv;
+    dev->vpp_timeout.expires = 0;
     req.Vpp1 = req.Vpp2 = 0;
     MTDHelperEntry(MTDSetVpp, link->handle, &req);
 }
@@ -343,8 +440,10 @@ static void vpp_shutdown(dev_link_t *link)
     flash_dev_t *dev;
     dev = (flash_dev_t *)link->priv;
     dev->vpp_usage--;
-    if (dev->vpp_usage == 0)
+    if (dev->vpp_usage == 0) {
+	dev->vpp_timeout.expires = vpp_timeout_period;
 	add_timer(&dev->vpp_timeout);
+    }
 }
 
 /*======================================================================
@@ -375,13 +474,6 @@ static void free_lock(dev_link_t *link)
     wake_up_interruptible(&link->pending);
 }
 
-/*====================================================================*/
-
-static void cs_error(int func, int ret)
-{
-    CardServices(ReportError, dev_info, (void *)func, (void *)ret);
-}
-
 /*======================================================================
 
     If we're linked into the kernel, this gets called from init.c.
@@ -417,7 +509,7 @@ static dev_link_t *flash_attach(void)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_attach()\n");
+	kmsg("flash_attach()");
 #endif
 
     /* Create new memory card device */
@@ -439,7 +531,8 @@ static dev_link_t *flash_attach(void)
     client_reg.EventMask =
 	CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
 	CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
-	CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
+	CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME |
+	CS_EVENT_READY_CHANGE;
     client_reg.event_handler = &flash_event;
     client_reg.Version = 0x0210;
     client_reg.event_callback_args.client_data = link;
@@ -470,7 +563,7 @@ static void flash_detach(dev_link_t *link)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_detach(0x%p)\n", link);
+	kmsg("flash_detach(0x%p)", link);
 #endif
     
     /* Locate device structure */
@@ -528,7 +621,7 @@ static void flash_config(dev_link_t *link)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_config(0x%p)\n", link);
+	kmsg("flash_config(0x%p)", link);
 #endif
 
     /* Allocate a 4K memory window */
@@ -610,7 +703,7 @@ static void flash_release(u_long arg)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_release(0x%p)\n", link);
+	kmsg("flash_release(0x%p)", link);
 #endif
 
     link->state &= ~DEV_CONFIG;
@@ -652,14 +745,17 @@ static int flash_read(dev_link_t *link, char *buf, mtd_request_t *req)
     flash_region_t *flash;
     region_info_t *region;
     mtd_mod_win_t mod;
-    u_long from, length, nb, time, cell;
+    u_long from, length, nb, cell;
     int ret;
+#ifdef PCMCIA_DEBUG
+    u_long time;
+#endif
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_read(0x%p, 0x%ld, 0x%p, 0x%lx, 0x%lx)\n",
-	       link, req->MediaID, buf, req->SrcCardOffset,
-	       req->TransferLength);
+	kmsg("flash_read(0x%p, 0x%ld, 0x%p, 0x%lx, 0x%lx)",
+	     link, req->MediaID, buf, req->SrcCardOffset,
+	     req->TransferLength);
 #endif
 
     flash = (flash_region_t *)(req->MediaID);
@@ -673,10 +769,18 @@ static int flash_read(dev_link_t *link, char *buf, mtd_request_t *req)
 	mod.Attributes = WIN_MEMORY_TYPE_CM;
     mod.AccessSpeed = region->AccessSpeed;
 
-    cell = (mod.CardOffset - region->CardOffset) / flash->cell_size;
-    ret = CS_SUCCESS;
     /* Suspend an in-progress block erase */
+    cell = (req->SrcCardOffset - region->CardOffset) / flash->cell_size;
     if (flash->cell[cell].state & FLASH_ERASING) {
+	if ((flash->cell[cell].erase_addr / region->BlockSize) ==
+	    (req->SrcCardOffset / region->BlockSize)) {
+#ifdef PCMCIA_DEBUG
+	    if (pc_debug > 1)
+		kmsg("delaying read...\n");
+#endif
+	    req->Status = MTD_WAITREQ;
+	    return CS_BUSY;
+	}
 	mod.CardOffset = flash->cell[cell].erase_addr;
 	ret = MTDHelperEntry(MTDModifyWindow, dev->ESRwin, &mod);
 	if (ret != CS_SUCCESS) goto done;
@@ -688,7 +792,10 @@ static int flash_read(dev_link_t *link, char *buf, mtd_request_t *req)
     mod.CardOffset = req->SrcCardOffset & ~(WINDOW_SIZE-1);
     from = req->SrcCardOffset & (WINDOW_SIZE-1);
     
-    time = jiffies;
+    ret = CS_SUCCESS;
+#ifdef PCMCIA_DEBUG
+    time = uticks();
+#endif
     for (length = req->TransferLength; length > 0; length -= nb) {
 	
 	ret = MTDHelperEntry(MTDModifyWindow, link->win, &mod);
@@ -706,10 +813,11 @@ static int flash_read(dev_link_t *link, char *buf, mtd_request_t *req)
     }
     
 #ifdef PCMCIA_DEBUG
-    time = jiffies - time;
-    if (pc_debug)
-	printk("read complete, time = %ld, avg = %ld ns\n",
-	       time, time*10000000/65536);
+    time = uticks() - time;
+    if (pc_debug && (time < 10000000))
+	kmsg("read complete, time = %ld, avg = %ld ns/word, rate = "
+	     "%ld kb/sec", time, time*2000/req->TransferLength,
+	     req->TransferLength*977/time);
 #endif
     
 done:
@@ -728,17 +836,32 @@ done:
     
 ======================================================================*/
 
-static int basic_write(char *esr, char *dest, char *buf,
-		       u_long nb, u_long is_krnl)
+static int basic_write(struct wait_queue **queue, char *esr, char *dest,
+		       char *buf, u_long nb, u_long is_krnl)
 {
     u_short npb;
     int ret;
 
+    /* Enable interrupts on write complete */
+    set_rdy_mode((u_short *)esr, IF_RDY_PULSE_WRITE);
+    
+    /* Fix for mis-aligned writes */
+    if ((u_long)dest & 1) {
+	kmsg("odd address fixup at 0x%p...", dest);
+	ret = page_setup(queue, (u_short *)esr, (u_short *)dest, 1);
+	if (ret != CS_SUCCESS) return ret;
+	*dest = (is_krnl) ? (*buf) : get_fs_byte(buf);
+	ret = page_write((u_short *)esr, (u_short *)dest, 1);
+	if (ret != CS_SUCCESS) return ret;
+	dest++; buf++; nb--;
+    }
+    
     for (; nb > 0; nb -= npb) {
 	    
 	/* npb = # of bytes to write to page buffer */
 	npb = (nb > 512) ? 512 : nb;
-	ret = page_setup((u_short *)esr, (u_short *)dest, npb);
+	/* sleep until page buffer is free */
+	ret = page_setup(queue, (u_short *)esr, (u_short *)dest, npb);
 	if (ret != CS_SUCCESS) return ret;
 	
 	if (is_krnl)
@@ -752,9 +875,11 @@ static int basic_write(char *esr, char *dest, char *buf,
 	dest += npb;
 	buf += npb;
     }
-    return check_write((u_short *)esr);
+    
+    /* sleep until block is ready */
+    return check_write(queue, (u_short *)esr);
 }
-
+    
 /*======================================================================
 
     The write request handler.  The Series 2+ cards support automatic
@@ -766,33 +891,52 @@ static int flash_write(dev_link_t *link, char *buf, mtd_request_t *req)
 {
     flash_dev_t *dev = (flash_dev_t *)link->priv;
     mtd_mod_win_t mod;
+    mtd_rdy_req_t rdy;
     flash_region_t *flash;
     region_info_t *region;
-    u_long from, length, nb, time, retry;
+    u_long from, length, nb, retry, cell;
     status_t status;
     int ret;
+#ifdef PCMCIA_DEBUG
+    u_long time;
+#endif
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_write(0x%p, 0x%lx, 0x%p, 0x%lx, 0x%lx)\n",
-	       link, req->MediaID, buf, req->DestCardOffset,
-	       req->TransferLength);
+	kmsg("flash_write(0x%p, 0x%lx, 0x%p, 0x%lx, 0x%lx)",
+	     link, req->MediaID, buf, req->DestCardOffset,
+	     req->TransferLength);
 #endif
 
     /* Check card write protect status */
-    if (CardServices(GetStatus, link->handle, &status) != CS_SUCCESS)
+    ret = CardServices(GetStatus, link->handle, &status);
+    if (ret != CS_SUCCESS) {
+	cs_error(GetStatus, ret);
 	return CS_GENERAL_FAILURE;
+    }
     if (status.CardState & CS_EVENT_WRITE_PROTECT)
 	return CS_WRITE_PROTECTED;
 
-    if (vpp_setup(link, req) != 0)
-	return CS_BUSY;
-
     flash = (flash_region_t *)(req->MediaID);
     region = &flash->region;
     if ((req->DestCardOffset / region->BlockSize) !=
 	((req->DestCardOffset+req->TransferLength-1) / region->BlockSize))
 	return CS_BAD_SIZE;
+    
+    if (vpp_setup(link, req) != 0)
+	return CS_BUSY;
+
+    /* Is this cell being erased or written? */
+    cell = (req->DestCardOffset - region->CardOffset) / flash->cell_size;
+    if (flash->cell[cell].state & FLASH_ERASING) {
+#ifdef PCMCIA_DEBUG
+	if (pc_debug > 1)
+	    kmsg("delaying write...");
+#endif
+	req->Status = MTD_WAITREQ;
+	return CS_BUSY;
+    }
+    
     if (region->Attributes & REGION_TYPE_AM)
 	mod.Attributes = WIN_MEMORY_TYPE_AM;
     else
@@ -803,10 +947,16 @@ static int flash_write(dev_link_t *link, char *buf, mtd_request_t *req)
     mod.CardOffset = req->DestCardOffset & ~(region->BlockSize-1);
     ret = MTDHelperEntry(MTDModifyWindow, dev->ESRwin, &mod);
     if (ret != CS_SUCCESS) goto done;
-    
-    time = jiffies;
+
+    rdy.Mask = CS_EVENT_READY_CHANGE;
+    MTDHelperEntry(MTDRDYMask, link->handle, &rdy);
+
+#ifdef PCMCIA_DEBUG
+    time = uticks();
+#endif
     mod.CardOffset = req->DestCardOffset & ~(WINDOW_SIZE-1);
     from = req->DestCardOffset & (WINDOW_SIZE-1);
+    link->state |= DEV_BUSY;
     
     for (length = req->TransferLength ; length > 0; length -= nb) {
 	ret = MTDHelperEntry(MTDModifyWindow, link->win, &mod);
@@ -814,17 +964,18 @@ static int flash_write(dev_link_t *link, char *buf, mtd_request_t *req)
 
 	nb = (from+length > WINDOW_SIZE) ? WINDOW_SIZE-from : length;
 
-	for (retry = 0; retry < write_limit; retry++) {
-	    ret = basic_write(dev->ESRbase, dev->Base+from, buf, nb,
+	for (retry = 0; retry < retry_limit; retry++) {
+	    ret = basic_write(&link->pending, dev->ESRbase,
+			      dev->Base+from, buf, nb,
 			      (req->Function & MTD_REQ_KERNEL));
 	    if (ret == CS_SUCCESS)
 		break;
 	    if (ret == CS_GENERAL_FAILURE)
 		goto done;
-	    reset_block((u_short *)dev->Base);
+	    reset_block((u_short *)dev->ESRbase);
 	}
-	if (retry == write_limit) {
-	    printk("write failed: too many retries!\n");
+	if (retry == retry_limit) {
+	    kmsg("write failed: too many retries!");
 	    goto done;
 	}
 	
@@ -834,16 +985,19 @@ static int flash_write(dev_link_t *link, char *buf, mtd_request_t *req)
     }
 
 #ifdef PCMCIA_DEBUG
-    time = jiffies - time;
-    if (pc_debug)
-	printk("write complete, time = %ld, avg = %ld us\n",
-	       time, time*10000/65536);
+    time = uticks() - time;
+    if (pc_debug && (time < 10000000))
+	kmsg("write complete, time = %ld, avg = %ld us/word, rate = "
+	     "%ld kb/sec", time, time*2/req->TransferLength,
+	     req->TransferLength*977/time);
 #endif
     
 done:
-    reset_block((u_short *)dev->Base);
+    reset_block((u_short *)dev->ESRbase);
+    rdy.Mask = 0;
+    MTDHelperEntry(MTDRDYMask, link->handle, &rdy);
+    link->state &= ~DEV_BUSY;
     /* Fire up the Vpp timer */
-    dev->vpp_timeout.expires = vpp_timeout_period;
     vpp_shutdown(link);
     return ret;
 } /* flash_write */
@@ -866,9 +1020,9 @@ static int flash_erase(dev_link_t *link, char *buf, mtd_request_t *req)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_erase(0x%p, 0x%lx, 0x%lx, 0x%lx)\n",
-	       link, req->MediaID, req->DestCardOffset,
-	       req->TransferLength);
+	kmsg("flash_erase(0x%p, 0x%lx, 0x%lx, 0x%lx)",
+	     link, req->MediaID, req->DestCardOffset,
+	     req->TransferLength);
 #endif
 
     flash = (flash_region_t *)(req->MediaID);
@@ -884,8 +1038,11 @@ static int flash_erase(dev_link_t *link, char *buf, mtd_request_t *req)
 	    return CS_BUSY;
 	}
 	/* Check card write protect status */
-	if (CardServices(GetStatus, link->handle, &status) != CS_SUCCESS)
+	ret = CardServices(GetStatus, link->handle, &status);
+	if (ret != CS_SUCCESS) {
+	    cs_error(GetStatus, ret);
 	    return CS_GENERAL_FAILURE;
+	}
 	if (status.CardState & CS_EVENT_WRITE_PROTECT)
 	    return CS_WRITE_PROTECTED;
 	flash->cell[i].state |= FLASH_PENDING;
@@ -910,22 +1067,35 @@ static int flash_erase(dev_link_t *link, char *buf, mtd_request_t *req)
 	flash->cell[i].state |= FLASH_ERASING;
 	flash->cell[i].erase_addr = mod.CardOffset;
 	flash->cell[i].erase_time = jiffies;
+	flash->cell[i].erase_retries = 0;
 	set_global_lock(dev->ESRwin, dev->ESRbase, 0);
+	/* Disable busy signal during the erase */
+	set_rdy_mode((u_short *)dev->Base, IF_RDY_DISABLE);
 	block_erase((u_short *)dev->Base);
     }
     else {
 	/* Check on an already started erase */
 	ret = check_erase((u_short *)dev->Base);
-	if (ret != CS_BUSY) {
-	    reset_block((u_short *)dev->Base);
+	if (ret == CS_SUCCESS)
 	    goto done;
+	else if (ret != CS_BUSY) {
+	    if (++flash->cell[i].erase_retries > retry_limit) {
+		kmsg("erase failed: too many retries!");
+		goto done;
+	    }
+	    else {
+		flash->cell[i].erase_time = jiffies;
+		reset_block((u_short *)dev->Base);
+		block_erase((u_short *)dev->Base);
+	    }
 	}
     }
 
     /* If the request is not complete, has it taken too long? */
     if (jiffies > flash->cell[i].erase_time + erase_limit) {
-	printk("erase timed out!\n");
+	kmsg("erase timed out!");
 	log_esr(NULL, (u_short *)dev->ESRbase);
+	abort_cmd((u_short *)dev->ESRbase);
 	ret = CS_GENERAL_FAILURE;
 	goto done;
     }
@@ -936,10 +1106,11 @@ static int flash_erase(dev_link_t *link, char *buf, mtd_request_t *req)
 done:
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("erase complete, time = %ld\n",
-	       jiffies - flash->cell[i].erase_time);
+	kmsg("erase complete, time = %ld",
+	     jiffies - flash->cell[i].erase_time);
 #endif
     flash->cell[i].state &= ~(FLASH_ERASING|FLASH_PENDING);
+    reset_block((u_short *)dev->ESRbase);
     set_global_lock(dev->ESRwin, dev->ESRbase, 1);
     vpp_shutdown(link);
     return ret;
@@ -952,6 +1123,11 @@ static int flash_request(dev_link_t *link, void *buf, mtd_request_t *req)
     int ret = 0;
     if (!(link->state & DEV_PRESENT))
 	return CS_NO_CARD;
+    /* Don't interfere with a write in progress */
+    if (link->state & DEV_BUSY) {
+	req->Status = MTD_WAITREQ;
+	return CS_BUSY;
+    }
     switch (req->Function & MTD_REQ_ACTION) {
     case MTD_REQ_READ:
 	ret = flash_read(link, buf, req);
@@ -987,7 +1163,7 @@ static int flash_event(event_t event, int priority,
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_event(%x)\n", event);
+	kmsg("flash_event(%x)", event);
 #endif
     
     switch (event) {
@@ -1004,6 +1180,10 @@ static int flash_event(event_t event, int priority,
 	link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
 	flash_config(link);
 	break;
+
+    case CS_EVENT_READY_CHANGE:
+	wake_up_interruptible(&link->pending);
+	break;
 	
     case CS_EVENT_PM_SUSPEND:
 	link->state |= DEV_SUSPEND;
@@ -1043,7 +1223,7 @@ int init_module(void)
     
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("iflash2+_mtd: Card Services release does not match!\n");
+	kmsg("Card Services release does not match!");
 	return -1;
     }
     
@@ -1054,7 +1234,7 @@ int init_module(void)
 
 void cleanup_module(void)
 {
-    printk("iflash2+_mtd: unloading\n");
+    kmsg("unloading");
     unregister_pcmcia_driver(&dev_info);
     while (dev_list != NULL)
 	flash_detach(dev_list);
