@@ -1,8 +1,13 @@
 /*======================================================================
 
-    A simple MTD for Intel Series 2 Flash devices
+    A simple MTD for Intel Series 2+ Flash devices
 
     Written by David Hinds, dhinds@allegro.stanford.edu
+
+    For efficiency and simplicity, this driver is very block oriented.
+    Reads and writes must not span erase block boundaries.  Erases
+    are limited to one erase block per request.  This makes it much
+    easier to manage multiple asynchronous erases efficiently.
     
 ======================================================================*/
 
@@ -39,17 +44,17 @@
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-"iflash2_mtd.c 1.5 1995/05/15 01:42:03 (David Hinds)\n";
+"iflash2_mtd.c 1.7 1995/05/27 04:42:36 (David Hinds)\n";
 #endif
 
-static int vpp_timeout_period	= 500;	/* in ticks */
+static int vpp_timeout_period	= 1000;	/* in ticks */
 static int vpp_settle		= 10;	/* in ticks */
 static int erase_timeout	= 100;	/* in ms */
-static int erase_limit		= 100;	/* times erase_timeout */
+static int erase_limit		= 1000;	/* in ticks */
 
-static u_short max_tries	= 4096;
+static int retry_limit		= 4;	/* max retries */
 
-#define USE_MEMCPY
+static u_long max_tries       	= 4096;
 
 /*====================================================================*/
 
@@ -64,18 +69,32 @@ static int flash_event(event_t event, int priority,
 static dev_link_t *flash_attach(void);
 static void flash_detach(dev_link_t *);
 
+#define MAX_CELLS		32
+
+/* A flash region is composed of one or more "cells", where we allow
+   simultaneous erases if they are in different cells */
+typedef struct flash_region_t {
+    region_info_t	region;
+    u_long		cell_size;
+    struct flash_cell_t {
+	u_long		state;
+	u_long		erase_time;
+	u_long		erase_addr;
+	u_long		erase_retries;
+    } cell[MAX_CELLS];
+} flash_region_t;
+
 typedef struct flash_dev_t {
-    u_long		state;
-    u_long		erase_start;
-    u_long		erase_retries;
     caddr_t		Base;
-    int			vpp_ok;
+    int			vpp_usage;
+    u_long		vpp_start;
     struct timer_list	vpp_timeout;
-    u_long		cur_erase;
-    region_info_t	region[2*CISTPL_MAX_DEVICES];
+    flash_region_t	*flash[2*CISTPL_MAX_DEVICES];
 } flash_dev_t;
 
-#define FLASH_ERASING		0x01
+#define FLASH_PENDING		0x01
+#define FLASH_ERASING		0x02
+#define FLASH_ERASE_SUSPEND	0x04
 
 static dev_info_t dev_info = "iflash2_mtd";
 
@@ -83,6 +102,28 @@ static dev_link_t *dev_list = NULL;
 
 /*====================================================================*/
 
+static void kmsg(char *fmt, ...)
+{
+    char buf[256];
+    va_list args;
+    
+    va_start(args, fmt);
+    vsprintf(buf, fmt, args);
+    va_end(args);
+    printk("%s: %s\n", (char *)dev_info, buf);
+}
+
+static void cs_error(int func, int ret)
+{
+    CardServices(ReportError, dev_info, (void *)func, (void *)ret);
+}
+
+/*======================================================================
+
+    Low level routines for programming the flash card.
+    
+======================================================================*/
+
 static inline int byte_write(volatile u_char *address, u_char data)
 {
     register u_char CSR = 0;
@@ -95,8 +136,8 @@ static inline int byte_write(volatile u_char *address, u_char data)
 	    return CS_SUCCESS;
 	}
     }
-    printk("byte_write timed out at 0x%p, CSR = 0x%x\n", address, CSR);
-    return CS_WRITE_FAILURE;
+    kmsg("byte_write timed out at 0x%p, CSR = 0x%x", address, CSR);
+    return CS_GENERAL_FAILURE;
 }
 
 static inline int word_write(volatile u_short *address, u_short data)
@@ -122,8 +163,8 @@ __asm__ __volatile__(
 	    return CS_SUCCESS;
 	}
     }
-    printk("word_write timed out at 0x%p, CSR = 0x%x\n", address, CSR);
-    return CS_WRITE_FAILURE;
+    kmsg("word_write timed out at 0x%p, CSR = 0x%x", address, CSR);
+    return CS_GENERAL_FAILURE;
 #endif
 }
 
@@ -136,11 +177,11 @@ static int check_write(volatile u_short *address)
 	if ((CSR & CSR_WR_READY) == CSR_WR_READY) break;
     }
     if (i == max_tries) {
-	printk("check_write: timed out!  CSR = 0x%x\n", CSR);
-	return CS_WRITE_FAILURE;
+	kmsg("check_write: timed out!  CSR = 0x%x", CSR);
+	return CS_GENERAL_FAILURE;
     }
     if (CSR & (CSR_WR_ERR | CSR_VPP_LOW)) {
-	printk("write error: CSR = 0x%x\n", CSR);
+	kmsg("write error: CSR = 0x%x", CSR);
 	return CS_WRITE_FAILURE;
     }
     else
@@ -161,13 +202,42 @@ static int check_erase(volatile u_short *address)
     if ((CSR & CSR_WR_READY) != CSR_WR_READY)
 	return CS_BUSY;
     else if (CSR & (CSR_ERA_ERR | CSR_VPP_LOW | CSR_WR_ERR)) {
-	printk("erase failed: CSR = 0x%x\n", CSR);
+	kmsg("erase failed: CSR = 0x%x", CSR);
 	return CS_WRITE_FAILURE;
     }
     else
 	return CS_SUCCESS;
 }
 
+static int suspend_erase(volatile u_short *address)
+{
+    u_short CSR = 0;
+    u_long i;
+
+    *address = IF_ERASE_SUSPEND;
+    *address = IF_READ_CSR;
+    for (i = 0; i < max_tries; i++) {
+	CSR = *address;
+	if ((CSR & CSR_WR_READY) == CSR_WR_READY) break;
+    }
+    if (i == max_tries) {
+	kmsg("suspend_erase timed out: CSR = 0x%x", CSR);
+	return CS_GENERAL_FAILURE;
+    }
+    *address = IF_READ_ARRAY;
+    return CS_SUCCESS;
+}
+
+static void resume_erase(volatile u_short *address)
+{
+    u_short CSR;
+    *address = IF_READ_CSR;
+    CSR = *address;
+    /* Only give resume signal if the erase is really suspended */
+    if (CSR & CSR_ERA_SUSPEND)
+	*address = IF_CONFIRM;
+}
+
 static void reset_block(volatile u_short *address)
 {
     *address = IF_CLEAR_CSR;
@@ -180,34 +250,50 @@ static void reset_block(volatile u_short *address)
     see if Vpp is available for the specified device.  If not, it
     turns on Vpp.  The vpp_shutdown() function is scheduled to turn
     Vpp off after an interval of inactivity.
+
+    vpp_setup() assumes that it will be called at the top of a
+    request handler, and that it can use the MTD_REQ_TIMEOUT flag
+    to tell if it has already been called for this particular
+    request, so that it can count Vpp users.
+
+    A handler should call vpp_shutdown() once for each request that
+    does a vpp_setup().
     
 ======================================================================*/
 
 static int vpp_setup(dev_link_t *link, mtd_request_t *req)
 {
     flash_dev_t *dev = (flash_dev_t *)link->priv;
-    if (dev->vpp_ok)
-	del_timer(&dev->vpp_timeout);
-    else {
-	if (req->Function & MTD_REQ_TIMEOUT)
-	    dev->vpp_ok = 1;
-	else {
-	    mtd_vpp_req_t vpp_req;
+    mtd_vpp_req_t vpp_req;
+
+    /* First time for this request? */
+    if (!(req->Function & MTD_REQ_TIMEOUT)) {
+	dev->vpp_usage++;
+	/* If no other users, kill shutdown timer and apply power */
+	if (dev->vpp_usage == 1) {
 #ifdef PCMCIA_DEBUG
 	    if (pc_debug > 1)
-		printk("Raising Vpp...\n");
+		kmsg("raising Vpp...");
 #endif
-	    vpp_req.Vpp1 = vpp_req.Vpp2 = 120;
-	    MTDHelperEntry(MTDSetVpp, link->handle, &vpp_req);
-	    req->Status = MTD_WAITTIMER;
-	    req->Timeout = vpp_settle;
-	    return 1;
+	    if (dev->vpp_timeout.expires)
+		del_timer(&dev->vpp_timeout);
+	    else {
+		dev->vpp_start = jiffies;
+		vpp_req.Vpp1 = vpp_req.Vpp2 = 120;
+		MTDHelperEntry(MTDSetVpp, link->handle, &vpp_req);
+	    }
 	}
     }
+    /* Wait for Vpp to settle if it was just applied */
+    if (jiffies < dev->vpp_start + vpp_settle) {
+	req->Status = MTD_WAITTIMER;
+	req->Timeout = vpp_settle * 10;
+	return 1;
+    }
     return 0;
 }
 
-static void vpp_shutdown(u_long arg)
+static void vpp_off(u_long arg)
 {
     dev_link_t *link = (dev_link_t *)arg;
     flash_dev_t *dev;
@@ -215,12 +301,23 @@ static void vpp_shutdown(u_long arg)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("Lowering Vpp...\n");
+	kmsg("lowering Vpp...");
 #endif
+    dev = (flash_dev_t *)link->priv;
+    dev->vpp_timeout.expires = 0;
     req.Vpp1 = req.Vpp2 = 0;
     MTDHelperEntry(MTDSetVpp, link->handle, &req);
+}
+
+static void vpp_shutdown(dev_link_t *link)
+{
+    flash_dev_t *dev;
     dev = (flash_dev_t *)link->priv;
-    dev->vpp_ok = 0;
+    dev->vpp_usage--;
+    if (dev->vpp_usage == 0) {
+	dev->vpp_timeout.expires = vpp_timeout_period;
+	add_timer(&dev->vpp_timeout);
+    }
 }
 
 /*======================================================================
@@ -251,13 +348,6 @@ static void free_lock(dev_link_t *link)
     wake_up_interruptible(&link->pending);
 }
 
-/*====================================================================*/
-
-static void cs_error(int func, int ret)
-{
-    CardServices(ReportError, dev_info, (void *)func, (void *)ret);
-}
-
 /*======================================================================
 
     If we're linked into the kernel, this gets called from init.c.
@@ -293,7 +383,7 @@ static dev_link_t *flash_attach(void)
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_attach()\n");
+	kmsg("flash_attach()");
 #endif
 
     /* Create new memory card device */
@@ -303,7 +393,7 @@ static dev_link_t *flash_attach(void)
     link->release.data = (u_long)link;
     dev = kmalloc(sizeof(struct flash_dev_t), GFP_KERNEL);
     memset(dev, 0, sizeof(*dev));
-    dev->vpp_timeout.function = vpp_shutdown;
+    dev->vpp_timeout.function = vpp_off;
     dev->vpp_timeout.data = (u_long)link;
     link->priv = dev;
 
@@ -346,7 +436,7 @@ static void flash_detach(dev_link_t *link)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_detach(0x%p)\n", link);
+	kmsg("flash_detach(0x%p)", link);
 #endif
     
     /* Locate device structure */
@@ -404,7 +494,7 @@ static void flash_config(dev_link_t *link)
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_config(0x%p)\n", link);
+	kmsg("flash_config(0x%p)", link);
 #endif
 
     /* Allocate a 4K memory window */
@@ -419,12 +509,12 @@ static void flash_config(dev_link_t *link)
 	flash_release((u_long)link);
 	return;
     }
+    dev = link->priv;
+    dev->Base = req.Base;
 
     link->state |= DEV_CONFIG;
 
     /* Grab info for all the memory regions we can access */
-    dev = link->priv;
-    dev->Base = req.Base;
     i = 0;
     for (attr = 0; attr < 2; attr++) {
 	region.Attributes = attr ? REGION_TYPE_AM : REGION_TYPE_CM;
@@ -432,17 +522,27 @@ static void flash_config(dev_link_t *link)
 	while (ret == CS_SUCCESS) {
 	    reg.Attributes = region.Attributes;
 	    reg.Offset = region.CardOffset;
-	    reg.MediaID = (u_long)&dev->region[i];
+	    dev->flash[i] = kmalloc(sizeof(struct flash_region_t),
+				    GFP_KERNEL);
+	    reg.MediaID = (u_long)dev->flash[i];
 	    ret = CardServices(RegisterMTD, link->handle, &reg);
-	    if (ret != 0) break;		
+	    if (ret != 0) {
+		kfree_s(dev->flash[i], sizeof(struct flash_region_t));
+		break;
+	    }
 	    printk("iflash2_mtd: %s at 0x%lx, ",
 		   attr ? "attr" : "common", region.CardOffset);
 	    printk_size(region.RegionSize);
 	    printk(", %ld ns\n", region.AccessSpeed);
-	    dev->region[i] = region; i++;
+	    memset(dev->flash[i], 0, sizeof(struct flash_region_t));
+	    dev->flash[i]->region = region;
+	    /* All Series 2 cards have 2MB component pairs */
+	    dev->flash[i]->cell_size = 0x200000;
+	    i++;
 	    ret = CardServices(GetNextRegion, link->handle, &region);
 	}
     }
+    dev->flash[i] = NULL;
     
 } /* flash_config */
 
@@ -457,116 +557,212 @@ static void flash_release(u_long arg)
 {
     dev_link_t *link = (dev_link_t *)arg;
     flash_dev_t *dev;
+    int i;
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_release(0x%p)\n", link);
+	kmsg("flash_release(0x%p)", link);
 #endif
 
+    link->state &= ~DEV_CONFIG;
     if (link->win) {
 	int ret = CardServices(ReleaseWindow, link->win);
 	if (ret != CS_SUCCESS)
 	    cs_error(ReleaseWindow, ret);
     }
     dev = link->priv;
-    if (dev->vpp_ok) {
+    if (dev->vpp_usage == 0)
 	del_timer(&dev->vpp_timeout);
-	vpp_shutdown((u_long)link);
-    }
-    link->state &= ~DEV_CONFIG;
+    vpp_off((u_long)link);
+    for (i = 0; i < 2*CISTPL_MAX_DEVICES; i++)
+	if (dev->flash[i])
+	    kfree_s(dev->flash[i], sizeof(struct flash_region_t));
+	else break;
     
     if (link->state & DEV_STALE_LINK)
 	flash_detach(link);
     
 } /* flash_release */
 
-/*====================================================================*/
+/*======================================================================
+
+    The read request handler.  This handler supports suspending
+    current erase requests.  Reading from a block that is currently
+    erasing is undefined.
+    
+======================================================================*/
 
 static int flash_read(dev_link_t *link, char *buf, mtd_request_t *req)
 {
     flash_dev_t *dev = (flash_dev_t *)link->priv;
+    flash_region_t *flash;
     region_info_t *region;
     mtd_mod_win_t mod;
-    u_long from, length, nb;
+    u_long from, length, nb, time, cell;
     int ret;
     
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_read(0x%p, 0x%ld, 0x%p, 0x%lx, 0x%lx)\n",
-	       link, req->MediaID, buf, req->SrcCardOffset,
-	       req->TransferLength);
+	kmsg("flash_read(0x%p, 0x%ld, 0x%p, 0x%lx, 0x%lx)",
+	     link, req->MediaID, buf, req->SrcCardOffset,
+	     req->TransferLength);
 #endif
 
-    region = (region_info_t *)(req->MediaID);
+    flash = (flash_region_t *)(req->MediaID);
+    region = &flash->region;
+    if ((req->SrcCardOffset / region->BlockSize) !=
+	((req->SrcCardOffset+req->TransferLength-1) / region->BlockSize))
+	return CS_BAD_SIZE;
     if (region->Attributes & REGION_TYPE_AM)
 	mod.Attributes = WIN_MEMORY_TYPE_AM;
     else
 	mod.Attributes = WIN_MEMORY_TYPE_CM;
     mod.AccessSpeed = region->AccessSpeed;
 
+    /* Suspend an in-progress block erase */
+    cell = (req->SrcCardOffset - region->CardOffset) / flash->cell_size;
+    if (flash->cell[cell].state & FLASH_ERASING) {
+	if ((flash->cell[cell].erase_addr / region->BlockSize) ==
+	    (req->SrcCardOffset / region->BlockSize)) {
+	    req->Status = MTD_WAITREQ;
+	    return CS_BUSY;
+	}
+	mod.CardOffset = flash->cell[cell].erase_addr;
+	ret = MTDHelperEntry(MTDModifyWindow, link->win, &mod);
+	if (ret != CS_SUCCESS) goto done;
+	ret = suspend_erase((u_short *)dev->Base);
+	if (ret != CS_SUCCESS) goto done;
+	flash->cell[cell].state |= FLASH_ERASE_SUSPEND;
+    }
+
     mod.CardOffset = req->SrcCardOffset & ~(WINDOW_SIZE-1);
     from = req->SrcCardOffset & (WINDOW_SIZE-1);
+    
+    ret = CS_SUCCESS;
+    time = jiffies;
     for (length = req->TransferLength; length > 0; length -= nb) {
-	ret = MTDHelperEntry(MTDModifyWindow, link->win, &mod);
-	if (ret != CS_SUCCESS) {
-	    cs_error(MapMemPage, ret);
-	    return ret;
-	}
-	nb = (from+length > WINDOW_SIZE) ? WINDOW_SIZE-from : length;
 	
-#ifdef USE_MEMCPY
+	ret = MTDHelperEntry(MTDModifyWindow, link->win, &mod);
+	if (ret != CS_SUCCESS) goto done;
+	nb = (from+length > WINDOW_SIZE) ? WINDOW_SIZE-from : length;
+
 	if (req->Function & MTD_REQ_KERNEL)
 	    memcpy(buf, &dev->Base[from], nb);
 	else
 	    memcpy_tofs(buf, &dev->Base[from], nb);
-	buf += nb;
-#else
-	if (req->Function & MTD_REQ_KERNEL)
-	    for (i = 0; i < nb; i++, buf++)
-		*buf = dev->Base[from+i];
-	else
-	    for (i = 0; i < nb; i++, buf++)
-		put_fs_byte(dev->Base[from+i], buf);
-#endif
 	
+	buf += nb;
 	from = 0;
 	mod.CardOffset += WINDOW_SIZE;
     }
-    return CS_SUCCESS;
+    
+#ifdef PCMCIA_DEBUG
+    time = jiffies - time;
+    if (pc_debug && (time > 1))
+	kmsg("read complete, time = %ld, avg = %ld ns/word, rate = "
+	     "%ld kb/sec", time, time*20000000/req->TransferLength,
+	     req->TransferLength*100/(time*1024));
+#endif
+    
+done:
+    if (flash->cell[cell].state & FLASH_ERASE_SUSPEND) {
+	mod.CardOffset = flash->cell[cell].erase_addr;
+	ret = MTDHelperEntry(MTDModifyWindow, link->win, &mod);
+	if (ret == CS_SUCCESS)
+	    resume_erase((u_short *)dev->Base);
+	flash->cell[cell].state &= ~FLASH_ERASE_SUSPEND;
+    }
+    return ret;
 } /* flash_read */
 
-/*====================================================================*/
+/*======================================================================
+
+    basic_write() handles a write that fits completely within a
+    memory window that has already been set up.
+    
+======================================================================*/
+
+static int basic_write(char *dest, char *buf,
+		       u_long nb, u_long is_krnl)
+{
+    int ret;
+    
+    *(u_short *)dest = IF_READ_CSR;
+    if (is_krnl) {
+	if (nb & 1) {
+	    ret = byte_write(dest, *buf);
+	    if (ret != CS_SUCCESS) return ret;
+		dest++; buf++; nb--;
+	}
+	for (; nb != 0; dest += 2, buf += 2, nb -= 2) {
+	    ret = word_write((u_short *)dest, *(u_short *)buf);
+	    if (ret != CS_SUCCESS) return ret;
+	}
+    }
+    else {
+	if (nb & 1) {
+	    ret = byte_write(dest, get_fs_byte(buf));
+	    if (ret != CS_SUCCESS) return ret;
+	    dest++; buf++; nb--;
+	}
+	for (; nb != 0; dest += 2, buf += 2, nb -= 2) {
+	    ret = word_write((u_short *)dest, get_fs_word((u_short *)buf));
+	    if (ret != CS_SUCCESS) return ret;
+	}
+    }
+    return check_write((u_short *)dest);
+	
+} /* basic_write */
+
+/*======================================================================
+
+    The write request handler.  The Series 2+ cards support automatic
+    erase suspend for writes.
+    
+======================================================================*/
 
 static int flash_write(dev_link_t *link, char *buf, mtd_request_t *req)
 {
     flash_dev_t *dev = (flash_dev_t *)link->priv;
     mtd_mod_win_t mod;
+    flash_region_t *flash;
     region_info_t *region;
-    u_long from, length, nb, i, time;
-    u_char *dest;
+    u_long from, length, nb, time, retry, cell;
     status_t status;
     int ret;
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_write(0x%p, 0x%lx, 0x%p, 0x%lx, 0x%lx)\n",
-	       link, req->MediaID, buf, req->DestCardOffset,
-	       req->TransferLength);
+	kmsg("flash_write(0x%p, 0x%lx, 0x%p, 0x%lx, 0x%lx)",
+	     link, req->MediaID, buf, req->DestCardOffset,
+	     req->TransferLength);
 #endif
 
     /* Check card write protect status */
     ret = CardServices(GetStatus, link->handle, &status);
-    if (ret != 0) {
+    if (ret != CS_SUCCESS) {
 	cs_error(GetStatus, ret);
 	return CS_GENERAL_FAILURE;
     }
     if (status.CardState & CS_EVENT_WRITE_PROTECT)
 	return CS_WRITE_PROTECTED;
 
+    flash = (flash_region_t *)(req->MediaID);
+    region = &flash->region;
+    if ((req->DestCardOffset / region->BlockSize) !=
+	((req->DestCardOffset+req->TransferLength-1) / region->BlockSize))
+	return CS_BAD_SIZE;
+    
     if (vpp_setup(link, req) != 0)
 	return CS_BUSY;
 
-    region = (region_info_t *)(req->MediaID);
+    /* Is this cell being erased? */
+    cell = (req->DestCardOffset - region->CardOffset) / flash->cell_size;
+    if (flash->cell[cell].state & FLASH_ERASING) {
+	req->Status = MTD_WAITREQ;
+	return CS_BUSY;
+    }
+    
     if (region->Attributes & REGION_TYPE_AM)
 	mod.Attributes = WIN_MEMORY_TYPE_AM;
     else
@@ -576,142 +772,136 @@ static int flash_write(dev_link_t *link, char *buf, mtd_request_t *req)
     time = jiffies;
     mod.CardOffset = req->DestCardOffset & ~(WINDOW_SIZE-1);
     from = req->DestCardOffset & (WINDOW_SIZE-1);
+    
     for (length = req->TransferLength ; length > 0; length -= nb) {
 	ret = MTDHelperEntry(MTDModifyWindow, link->win, &mod);
-	if (ret != CS_SUCCESS) {
-	    cs_error(MapMemPage, ret);
+	if (ret != CS_SUCCESS) goto done;
+
+	nb = (from+length > WINDOW_SIZE) ? WINDOW_SIZE-from : length;
+
+	for (retry = 0; retry < retry_limit; retry++) {
+	    ret = basic_write(dev->Base+from, buf, nb,
+			      (req->Function & MTD_REQ_KERNEL));
+	    if (ret == CS_SUCCESS)
+		break;
+	    if (ret == CS_GENERAL_FAILURE)
+		goto done;
+	    reset_block((u_short *)dev->Base);
+	}
+	if (retry == retry_limit) {
+	    kmsg("write failed: too many retries!");
 	    goto done;
 	}
-	nb = (from+length > WINDOW_SIZE) ? WINDOW_SIZE-from : length;
-	
-	dest = dev->Base+from;
-	*(u_short *)dest = IF_READ_CSR;
-	if (req->Function & MTD_REQ_KERNEL) {
-	    i = nb;
-	    if (i & 1) {
-		ret = byte_write(dest, *buf);
-		if (ret != CS_SUCCESS) goto done;
-		dest++; buf++; i--;
-	    }
-	    for (; i != 0; dest += 2, buf += 2, i -= 2) {
-		ret = word_write((u_short *)dest, *(u_short *)buf);
-		if (ret != CS_SUCCESS) goto done;
-	    }
-	}
-	else {
-	    i = nb;
-	    if (i & 1) {
-		ret = byte_write(dest, get_fs_byte(buf));
-		if (ret != CS_SUCCESS) goto done;
-		dest++; buf++; i--;
-	    }
-	    for (; i != 0; dest += 2, buf += 2, i -= 2) {
-		ret = word_write((u_short *)dest, get_fs_word((u_short *)buf));
-		if (ret != CS_SUCCESS) goto done;
-	    }
-	}
-	ret = check_write((u_short *)dev->Base);
-	if (ret != CS_SUCCESS) goto done;
 	
+	buf += nb;
 	from = 0;
 	mod.CardOffset += WINDOW_SIZE;
     }
 
-done:
 #ifdef PCMCIA_DEBUG
     time = jiffies - time;
-    if (pc_debug)
-	printk("write complete, time = %ld, avg = %ld us\n",
-	       time, time*10000/65536);
+    if (pc_debug && (time > 1))
+	kmsg("write complete, time = %ld, avg = %ld us/word, rate = "
+	     "%ld kb/sec", time, time*20000/req->TransferLength,
+	     req->TransferLength*100/(time*1024));
 #endif
+    
+done:
     reset_block((u_short *)dev->Base);
     /* Fire up the Vpp timer */
     dev->vpp_timeout.expires = vpp_timeout_period;
-    add_timer(&dev->vpp_timeout);
+    vpp_shutdown(link);
     return ret;
 } /* flash_write */
 
-/*====================================================================*/
+/*======================================================================
+
+    The erase request handler.  This handler supports simultaneous
+    erases in different device components.
+    
+======================================================================*/
 
 static int flash_erase(dev_link_t *link, char *buf, mtd_request_t *req)
 {
     flash_dev_t *dev = (flash_dev_t *)link->priv;
     status_t status;
+    flash_region_t *flash;
     region_info_t *region;
     mtd_mod_win_t mod;
-    volatile u_short *addr;
-    int ret;
+    int i, ret;
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_erase(0x%p, 0x%lx, 0x%lx, 0x%lx)\n",
-	       link, req->MediaID, req->DestCardOffset,
-	       req->TransferLength);
+	kmsg("flash_erase(0x%p, 0x%lx, 0x%lx, 0x%lx)",
+	     link, req->MediaID, req->DestCardOffset,
+	     req->TransferLength);
 #endif
 
-    /* First see if another erase is in progress */
-    if ((dev->state & FLASH_ERASING) &&
-	!(req->Function & MTD_REQ_TIMEOUT)) {
-	req->Status = MTD_WAITREQ;
-	return CS_BUSY;
-    }
+    flash = (flash_region_t *)(req->MediaID);
+    region = &flash->region;
+    if (region->BlockSize != req->TransferLength)
+	return CS_BAD_SIZE;
     
-    /* Check card write protect status */
-    ret = CardServices(GetStatus, link->handle, &status);
-    if (ret != 0) {
-	cs_error(GetStatus, ret);
-	return CS_GENERAL_FAILURE;
-    }
-    if (status.CardState & CS_EVENT_WRITE_PROTECT)
-	return CS_WRITE_PROTECTED;
+    i = (req->DestCardOffset-region->CardOffset)/flash->cell_size;
     
-    if (vpp_setup(link, req) != 0)
-	return CS_BUSY;
+    if (!(req->Function & MTD_REQ_TIMEOUT)) {
+	if (flash->cell[i].state & (FLASH_ERASING|FLASH_PENDING)) {
+	    req->Status = MTD_WAITREQ;
+	    return CS_BUSY;
+	}
+	/* Check card write protect status */
+	ret = CardServices(GetStatus, link->handle, &status);
+	if (ret != CS_SUCCESS) {
+	    cs_error(GetStatus, ret);
+	    return CS_GENERAL_FAILURE;
+	}
+	if (status.CardState & CS_EVENT_WRITE_PROTECT)
+	    return CS_WRITE_PROTECTED;
+	flash->cell[i].state |= FLASH_PENDING;
+	/* Activate Vpp if necessary */
+	if (vpp_setup(link, req) != 0)
+	    return CS_BUSY;
+    }
 
-    region = (region_info_t *)(req->MediaID);
     if (region->Attributes & REGION_TYPE_AM)
 	mod.Attributes = WIN_MEMORY_TYPE_AM;
     else
 	mod.Attributes = WIN_MEMORY_TYPE_CM;
     mod.AccessSpeed = region->AccessSpeed;
-
-    addr = (u_short *)dev->Base;
-    if (dev->state & FLASH_ERASING) {
-	mod.CardOffset = dev->cur_erase;
-	ret = MTDHelperEntry(MTDModifyWindow, link->win, &mod);
-	if (ret != CS_SUCCESS) {
-	    cs_error(MapMemPage, ret);
-	    goto done;
-	}
-	ret = check_erase(addr);
-	if (ret == CS_BUSY)
-	    goto retry;
-	else if (ret != CS_SUCCESS)
-	    goto done;
-	dev->cur_erase += region->BlockSize;
-	if (dev->cur_erase >= req->DestCardOffset+req->TransferLength)
-	    goto done;
+    mod.CardOffset = req->DestCardOffset;
+    ret = MTDHelperEntry(MTDModifyWindow, link->win, &mod);
+    if (ret != CS_SUCCESS)
+	goto done;
+    
+    if (flash->cell[i].state & FLASH_PENDING) {
+	/* Start a new block erase */
+	flash->cell[i].state &= ~FLASH_PENDING;
+	flash->cell[i].state |= FLASH_ERASING;
+	flash->cell[i].erase_addr = mod.CardOffset;
+	flash->cell[i].erase_time = jiffies;
+	block_erase((u_short *)dev->Base);
     }
     else {
-	dev->cur_erase = req->DestCardOffset;
-	dev->state |= FLASH_ERASING;
+	/* Check on an already started erase */
+	ret = check_erase((u_short *)dev->Base);
+	if (ret == CS_SUCCESS)
+	    goto done;
+	else if (ret != CS_BUSY) {
+	    if (++flash->cell[i].erase_retries > retry_limit) {
+		kmsg("erase failed: too many retries!");
+		goto done;
+	    }
+	    else {
+		flash->cell[i].erase_time = jiffies;
+		reset_block((u_short *)dev->Base);
+		block_erase((u_short *)dev->Base);
+	    }
+	}
     }
 
-    /* Start a new block erase */
-    mod.CardOffset = dev->cur_erase;
-    ret = MTDHelperEntry(MTDModifyWindow, link->win, &mod);
-    if (ret != CS_SUCCESS) {
-	cs_error(MapMemPage, ret);
-	goto done;
-    }
-
-    dev->erase_start = jiffies;
-    dev->erase_retries = 0;
-    block_erase(addr);
-
-retry:
-    if (++dev->erase_retries > erase_limit) {
-	printk("erase timed out!\n");
+    /* If the request is not complete, has it taken too long? */
+    if (jiffies > flash->cell[i].erase_time + erase_limit) {
+	kmsg("erase timed out!");
 	ret = CS_GENERAL_FAILURE;
 	goto done;
     }
@@ -722,13 +912,11 @@ retry:
 done:
 #ifdef PCMCIA_DEBUG
     if (pc_debug)
-	printk("erase complete, time = %ld\n", jiffies - dev->erase_start);
+	kmsg("erase complete, time = %ld",
+	     jiffies - flash->cell[i].erase_time);
 #endif
-    dev->state &= ~FLASH_ERASING;
-    reset_block(addr);
-    /* Fire up the Vpp timer */
-    dev->vpp_timeout.expires = vpp_timeout_period;
-    add_timer(&dev->vpp_timeout);
+    flash->cell[i].state &= ~(FLASH_ERASING|FLASH_PENDING);
+    vpp_shutdown(link);
     return ret;
 } /* flash_erase */
     
@@ -774,7 +962,7 @@ static int flash_event(event_t event, int priority,
 
 #ifdef PCMCIA_DEBUG
     if (pc_debug > 1)
-	printk("flash_event(%x)\n", event);
+	kmsg("flash_event(%x)", event);
 #endif
     
     switch (event) {
@@ -791,7 +979,7 @@ static int flash_event(event_t event, int priority,
 	link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
 	flash_config(link);
 	break;
-	
+
     case CS_EVENT_PM_SUSPEND:
 	link->state |= DEV_SUSPEND;
 	/* Fall through... */
@@ -830,7 +1018,7 @@ int init_module(void)
     
     CardServices(GetCardServicesInfo, &serv);
     if (serv.Revision != CS_RELEASE_CODE) {
-	printk("iflash2_mtd: Card Services release does not match!\n");
+	kmsg("Card Services release does not match!");
 	return -1;
     }
     
@@ -841,7 +1029,7 @@ int init_module(void)
 
 void cleanup_module(void)
 {
-    printk("iflash2_mtd: unloading\n");
+    kmsg("unloading");
     unregister_pcmcia_driver(&dev_info);
     while (dev_list != NULL)
 	flash_detach(dev_list);
