@@ -26,6 +26,7 @@
 #include <linux/timer.h>
 #include <linux/major.h>
 #include <linux/fs.h>
+#include <linux/ioctl.h>
 #include <asm/io.h>
 #include <asm/system.h>
 #include <asm/segment.h>
@@ -37,6 +38,7 @@
 #include "bulkmem.h"
 #include "cistpl.h"
 #include "ds.h"
+#include "memory.h"
 
 /* Major device #'s for memory device */
 static int major_dev = 0;
@@ -57,7 +59,7 @@ static int major_dev = 0;
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-"memory_cs.c 1.6 1995/05/15 01:43:40 (David Hinds)\n";
+"memory_cs.c 1.7 1995/05/23 19:46:29 (David Hinds)\n";
 #endif
 
 /*====================================================================*/
@@ -89,6 +91,7 @@ typedef struct minor_dev_t {
 typedef struct memory_dev_t {
     eraseq_handle_t	eraseq_handle;
     eraseq_entry_t	eraseq[MAX_ERASE];
+    struct wait_queue	*erase_pending;
     minor_dev_t		minor[2*CISTPL_MAX_DEVICES];
 } memory_dev_t;
 
@@ -115,7 +118,7 @@ static struct file_operations memory_chr_fops = {
     memory_write,      	/* write */
     NULL,		/* readdir */
     NULL,		/* select */
-    NULL,		/* ioctl */
+    memory_ioctl,	/* ioctl */
     NULL,		/* mmap */
     memory_open,	/* open */
     memory_close,	/* release */
@@ -419,6 +422,7 @@ static int memory_event(event_t event, int priority,
 		       event_callback_args_t *args)
 {
     dev_link_t *link = args->client_data;
+    memory_dev_t *dev;
     eraseq_entry_t *erase;
 
 #ifdef PCMCIA_DEBUG
@@ -446,7 +450,9 @@ static int memory_event(event_t event, int priority,
 	break;
     case CS_EVENT_ERASE_COMPLETE:
 	erase = (eraseq_entry_t *)(args->info);
-	wake_up_interruptible((struct wait_queue **)&erase->Optional);
+	wake_up((struct wait_queue **)&erase->Optional);
+	dev = (memory_dev_t *)(link->priv);
+	wake_up_interruptible(&dev->erase_pending);
 	break;
     case CS_EVENT_PM_SUSPEND:
 	link->state |= DEV_SUSPEND;
@@ -572,6 +578,50 @@ static int memory_read(struct inode *inode, struct file *file,
 	return -EIO;
 } /* memory_read */
 
+/*======================================================================
+
+    Erase a memory region.  This is used by the write routine for
+    suitably aligned and sized blocks.  It is also used for the
+    MEMERASE ioctl().
+    
+======================================================================*/
+
+static int memory_erase(memory_dev_t *dev, minor_dev_t *minor_dev,
+			u_long f_pos, int count)
+{
+    int i, ret;
+
+    /* Find a free erase slot, or wait for one to become available */
+    for (;;) {
+	for (i = 0; i < MAX_ERASE; i++)
+	    if (!ERASE_IN_PROGRESS(dev->eraseq[i].State)) break;
+	if (i < MAX_ERASE) break;
+	printk("waiting for erase slot...\n");
+	interruptible_sleep_on(&dev->erase_pending);
+	if (current->signal & ~current->blocked)
+	    return -ERESTARTSYS;
+    }
+
+    /* Queue a new request */
+    dev->eraseq[i].State = ERASE_QUEUED;
+    dev->eraseq[i].Handle = minor_dev->handle;
+    dev->eraseq[i].Offset = f_pos;
+    dev->eraseq[i].Size = count;
+    ret = CardServices(CheckEraseQueue, dev->eraseq_handle);
+    if (ret != CS_SUCCESS) {
+	cs_error(CheckEraseQueue, ret);
+	return -EIO;
+    }
+
+    /* Wait for request to complete */
+    dev->eraseq[i].Optional = NULL;
+    if (ERASE_IN_PROGRESS(dev->eraseq[i].State))
+	sleep_on((struct wait_queue **)&dev->eraseq[i].Optional);
+    if (dev->eraseq[i].State != ERASE_PASSED)
+	return -EIO;
+    return 0;
+}
+
 /*======================================================================
 
     Write for character-mode device
@@ -585,7 +635,7 @@ static int memory_write(struct inode *inode, struct file *file,
     memory_dev_t *dev;
     minor_dev_t *minor_dev;
     mem_op_t req;
-    int i, ret;
+    int ret;
 
     dev = (memory_dev_t *)dev_table[DEVICE_NR(minor)]->priv;
     minor_dev = &dev->minor[REGION_NR(minor)];
@@ -596,26 +646,9 @@ static int memory_write(struct inode *inode, struct file *file,
 #endif
     if (((file->f_pos & (minor_dev->region.BlockSize-1)) == 0) &&
 	((count & (minor_dev->region.BlockSize-1)) == 0)) {
-	for (i = 0; i < MAX_ERASE; i++)
-	    if (!ERASE_IN_PROGRESS(dev->eraseq[i].State)) break;
-	dev->eraseq[i].State = ERASE_QUEUED;
-	dev->eraseq[i].Handle = minor_dev->handle;
-	dev->eraseq[i].Offset = file->f_pos;
-	dev->eraseq[i].Size = count;
-	ret = CardServices(CheckEraseQueue, dev->eraseq_handle);
-	if (ret != CS_SUCCESS) {
-	    cs_error(CheckEraseQueue, ret);
-	    return -EIO;
-	}
-	dev->eraseq[i].Optional = NULL;
-	if (ERASE_IN_PROGRESS(dev->eraseq[i].State)) {
-	    interruptible_sleep_on((struct wait_queue **)
-				   &dev->eraseq[i].Optional);
-	    if (current->signal & ~current->blocked)
-		return -EINTR;
-	}
-	if (dev->eraseq[i].State != ERASE_PASSED)
-	    return -EIO;
+	ret = memory_erase(dev, minor_dev, file->f_pos, count);
+	if (ret != 0)
+	    return ret;
     }
     
     req.Attributes = 0;
@@ -643,6 +676,8 @@ static int memory_ioctl(struct inode *inode, struct file *file,
     dev_link_t *link;
     memory_dev_t *dev;
     minor_dev_t *minor_dev;
+    erase_info_t erase;
+    u_int size;
     int ret;
 
     link = dev_table[DEVICE_NR(minor)];
@@ -651,13 +686,32 @@ static int memory_ioctl(struct inode *inode, struct file *file,
 	return -ENODEV;
     dev = (memory_dev_t *)link->priv;
     minor_dev = &dev->minor[REGION_NR(minor)];
+
+    size = (cmd & IOCSIZE_MASK) >> IOCSIZE_SHIFT;
+    if (cmd & IOC_IN) {
+	ret = verify_area(VERIFY_READ, (char *)arg, size);
+	if (ret) return ret;
+    }
+    if (cmd & IOC_OUT) {
+	ret = verify_area(VERIFY_WRITE, (char *)arg, size);
+	if (ret) return ret;
+    }
+
+    ret = 0;
     switch (cmd) {
     case BLKGETSIZE:
-	ret = verify_area(VERIFY_WRITE, (long *)arg, sizeof(long));
-	if (ret) break;
 	put_fs_long(minor_dev->region.RegionSize/SECTOR_SIZE,
 		    (long *)arg);
 	break;
+    case MEMGETINFO:
+	memcpy_tofs((region_info_t *)arg, &minor_dev->region,
+		    sizeof(struct region_info_t));
+	break;
+    case MEMERASE:
+	memcpy_fromfs(&erase, (erase_info_t *)arg,
+		      sizeof(struct erase_info_t));
+	ret = memory_erase(dev, minor_dev, erase.Offset, erase.Size);
+	break;
     default:
 	ret = -EINVAL;
     }
