@@ -1,11 +1,13 @@
 /* ----------------------------------------------------------------------------
 Linux PCMCIA ethernet adapter driver for the New Media Ethernet LAN.
+  nmclan_cs.c,v 0.14 1995/05/23 03:19:30 rpao Exp
 
   The Ethernet LAN uses the Advanced Micro Devices (AMD) Am79C940 Media
   Access Controller for Ethernet (MACE).  It is essentially the Am2150
   PCMCIA Ethernet card contained in the the Am2150 Demo Kit.
 
 Written by Roger C. Pao <rpao@paonet.org>
+  Copyright 1995 Roger C. Pao
 
   This software may be used and distributed according to the terms of
   the GNU Public License.
@@ -15,9 +17,11 @@ References
   Am2150 Technical Reference Manual
   Am79C940 (MACE) Data Sheet
   Am79C90 (C-LANCE) Data Sheet
+  Linux PCMCIA Programmer's Guide v1.17
 
   Eric Mears, New Media Corporation
   Tom Pollard, New Media Corporation
+  Dean Siasoyco, New Media Corporation
   Ken Lesniak, Silicon Graphics, Inc. <lesniak@boston.sgi.com>
   Donald Becker <becker@cesdis1.gsfc.nasa.gov>
   David Hinds <dhinds@allegro.stanford.edu>
@@ -41,26 +45,50 @@ Driver Notes and Issues
 -------------------------------------------------------------------------------
 
 1. Developed on a Dell 320SLi
-   PCMCIA Card Services 2.5.6
+   PCMCIA Card Services 2.6.0
    Linux dell 1.2.5 #6 Sat Apr 15 18:51:01 PDT 1995 i386
 
 2. rc.pcmcia may require loading pcmcia_core with io_speed=300:
    'insmod pcmcia_core.o io_speed=300'.
-   This will avoid problems with fast systems thinking the card is an
-   anonymous memory card.
+   This will avoid problems with fast systems which causes rx_framecnt
+   to return random values.
 
 3. Future: Allow hot extraction.  In the meantime, use 'ifconfig eth0 down'
    before extraction.
 
-4. Future: Multiple transmit request processing.
-
-5. Future: Multicast processing.
+4. Future: Multicast processing.
 
 
 -------------------------------------------------------------------------------
 History
 -------------------------------------------------------------------------------
+nmclan_cs.c,v
+ * Revision 1.5  1995/06/01  16:51:08  root
+ * *** empty log message ***
+ *
+ * Revision 0.14  1995/05/23  03:19:30  rpao
+ * Added, in nmclan_config(), "tuple.Attributes = 0;".
+ * Modified MACE ID check to ignore chip revision level.
+ * Avoid tx_free_frames race condition between _start_xmit and _interrupt.
+ *
+ * Revision 0.13  1995/05/18  05:56:34  rpao
+ * Statistics changes.
+ * Bug fix: nmclan_reset did not enable TX and RX: call restore_multicast_list.
+ * Bug fix: mace_interrupt checks ~MACE_IMR_DEFAULT.  Fixes driver lockup.
+ *
+ * Revision 0.12  1995/05/14  00:12:23  rpao
+ * Statistics overhaul.
+ *
+
+95/05/13 rpao	V0.10a
+		Bug fix: MACE statistics counters used wrong I/O ports.
+		Bug fix: mace_interrupt() needed to allow statistics to be
+		processed without RX or TX interrupts pending.
+95/05/11 rpao	V0.10
+		Multiple transmit request processing.
+		Modified statistics to use MACE counters where possible.
 95/05/10 rpao	V0.09 Bug fix: Must use IO_DATA_PATH_WIDTH_AUTO.
+		*Released
 95/05/10 rpao	V0.08
 		Bug fix: Make all non-exported functions private by using
 		static keyword.
@@ -68,7 +96,6 @@ History
 95/05/10 rpao	V0.07 Statistics.
 95/05/09 rpao	V0.06 Fix rx_framecnt problem by addition of PCIC wait states.
 
-rpao=Roger C. Pao <rpao@paonet.org>
 ---------------------------------------------------------------------------- */
 
 
@@ -76,7 +103,9 @@ rpao=Roger C. Pao <rpao@paonet.org>
 Conditional Compilation
 ---------------------------------------------------------------------------- */
 
-#define FORCE_I365_WAIT			0
+#define MULTI_TX			1
+#define TIMEOUT_TX			0
+#define RESET_ON_TIMEOUT		0
 
 
 /* ----------------------------------------------------------------------------
@@ -111,6 +140,7 @@ Include Files
 
 #include "cs_types.h"
 #include "cs.h"
+#include "cisreg.h"
 #include "cistpl.h"
 #include "ds.h"
 
@@ -121,6 +151,21 @@ Defines
 
 #define ETHER_ADDR_LEN			ETH_ALEN
 					/* 6 bytes in an Ethernet Address */
+#define MACE_LADRF_LEN			8
+					/* 8 bytes in Logical Address Filter */
+
+/* Loop Control Defines */
+#define MACE_MAX_IR_ITERATIONS		10
+#define MACE_MAX_RX_ITERATIONS		1
+	/*
+	TBD: Dean brought this up, and I assumed the hardware would
+	handle it:
+
+	If MACE_MAX_RX_ITERATIONS is > 1, rx_framecnt may still be
+	non-zero when the isr exits.  We may not get another interrupt
+	to process the remaining packets for some time.
+	*/
+
 
 /*
 The Am2150 has a Xilinx XC3042 field programmable gate array (FPGA)
@@ -128,6 +173,8 @@ which manages the interface between the MACE and the PCMCIA bus.  It
 also includes buffer management for the 32K x 8 SRAM to control up to
 four transmit and 12 receive frames at a time.
 */
+#define AM2150_MAX_TX_FRAMES		4
+#define AM2150_MAX_RX_FRAMES		12
 
 /* Am2150 Ethernet Card I/O Mapping */
 #define AM2150_RCV			0x00
@@ -173,6 +220,9 @@ four transmit and 12 receive frames at a time.
 #define MACE_RTR2			31
 
 /* MACE Bit Masks */
+#define MACE_XMTRC_EXDEF		0x80
+#define MACE_XMTRC_XMTRC		0x0F
+
 #define MACE_XMTFS_XMTSV		0x80
 #define MACE_XMTFS_UFLO			0x40
 #define MACE_XMTFS_LCOL			0x20
@@ -209,6 +259,15 @@ four transmit and 12 receive frames at a time.
 #define MACE_MACCC_ENXMT		0x02
 #define MACE_MACCC_ENRCV		0x01
 
+#define MACE_PHYCC_LNKFL		0x80
+#define MACE_PHYCC_DLNKTST		0x40
+#define MACE_PHYCC_REVPOL		0x20
+#define MACE_PHYCC_DAPC			0x10
+#define MACE_PHYCC_LRT			0x08
+#define MACE_PHYCC_ASEL			0x04
+#define MACE_PHYCC_RWAKE		0x02
+#define MACE_PHYCC_AWAKE		0x01
+
 #define MACE_IAC_ADDRCHG		0x80
 #define MACE_IAC_PHYADDR		0x04
 #define MACE_IAC_LOGADDR		0x02
@@ -229,15 +288,79 @@ four transmit and 12 receive frames at a time.
 
 
 #define MACE_IMR_DEFAULT \
-  (0xFF-(MACE_IR_CERR | MACE_IR_MPCO | MACE_IR_RCVINT | MACE_IR_XMTINT))
+  (0xFF - \
+    ( \
+      MACE_IR_CERR | \
+      MACE_IR_RCVCCO | \
+      MACE_IR_RNTPCO | \
+      MACE_IR_MPCO | \
+      MACE_IR_RCVINT | \
+      MACE_IR_XMTINT \
+    ) \
+  )
+#undef MACE_IMR_DEFAULT
+#define MACE_IMR_DEFAULT 0x00 /* New statistics handling: grab everything */
 
 
 /* ----------------------------------------------------------------------------
 Type Definitions
 ---------------------------------------------------------------------------- */
 
+typedef struct _mace_statistics {
+  /* MACE_XMTFS */
+  int xmtsv;
+  int uflo;
+  int lcol;
+  int more;
+  int one;
+  int defer;
+  int lcar;
+  int rtry;
+
+  /* MACE_XMTRC */
+  int exdef;
+  int xmtrc;
+
+  /* RFS1--Receive Status (RCVSTS) */
+  int oflo;
+  int clsn;
+  int fram;
+  int fcs;
+
+  /* RFS2--Runt Packet Count (RNTPC) */
+  int rfs_rntpc;
+
+  /* RFS3--Receive Collision Count (RCVCC) */
+  int rfs_rcvcc;
+
+  /* MACE_IR */
+  int jab;
+  int babl;
+  int cerr;
+  int rcvcco;
+  int rntpco;
+  int mpco;
+
+  /* MACE_MPC */
+  int mpc;
+
+  /* MACE_RNTPC */
+  int rntpc;
+
+  /* MACE_RCVCC */
+  int rcvcc;
+} mace_statistics;
+
+
 typedef struct _mace_private {
-    struct enet_statistics stats;
+  struct enet_statistics linux_stats; /* Linux statistics counters */
+  mace_statistics mace_stats; /* MACE chip statistics counters */
+
+  /* restore_multicast_list() state variables */
+  int multicast_ladrf[MACE_LADRF_LEN]; /* Logical address filter */
+  int multicast_num_addrs;
+
+  char tx_free_frames; /* Number of free transmit frame buffers */
 } mace_private;
 
 
@@ -245,7 +368,8 @@ typedef struct _mace_private {
 Private Global Variables
 ---------------------------------------------------------------------------- */
 
-static char *version="nmclan_cs 0.09 1995/05/11 (Roger C. Pao)\n";
+static char rcsid[] = "nmclan_cs.c,v 1.5 1995/06/01 16:51:08 root Exp\n";
+static char *version="nmclan_cs 1.5 (Roger C. Pao)\n";
 
 static dev_info_t dev_info="nmclan_cs";
 static dev_link_t *dev_list=NULL;
@@ -258,10 +382,12 @@ static char *if_names[]={
   "Auto"
 };
 
+
 #ifdef PCMCIA_DEBUG
 static int pc_debug=PCMCIA_DEBUG;
 #endif
 
+/* #define NMCLAN_DEBUG 6 */
 #ifdef NMCLAN_DEBUG
 static int nmclan_debug=NMCLAN_DEBUG;
 #else
@@ -307,6 +433,7 @@ static int mace_start_xmit(struct sk_buff *skb, struct device *dev);
 static void mace_interrupt(int irq, struct pt_regs *regs);
 static struct enet_statistics *mace_get_stats(struct device *dev);
 static int mace_rx(struct device *dev, unsigned char RxCnt);
+static void restore_multicast_list(struct device *dev);
 static void set_multicast_list(struct device *dev, int num_addrs, void *addrs);
 
 static dev_link_t *nmclan_attach(void);
@@ -372,6 +499,10 @@ static dev_link_t *nmclan_attach(void)
     printk("nmclan_attach()\n");
 #endif
 
+  if (nmclan_debug > 3)
+    printk(rcsid);
+
+
   /* Create new ethernet device */
   link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
   memset(link, 0, sizeof(struct dev_link_t));
@@ -395,6 +526,7 @@ static dev_link_t *nmclan_attach(void)
   /* Allocate private data area for this device. */
   dev->priv = kmalloc(sizeof(mace_private), GFP_KERNEL);
   memset(dev->priv, 0, sizeof(mace_private));
+  ((mace_private *)dev->priv)->tx_free_frames=AM2150_MAX_TX_FRAMES;
 
   dev->hard_start_xmit = &mace_start_xmit;
   dev->set_config = &mace_config;
@@ -596,6 +728,7 @@ static void mace_init(int ioaddr, char *enet_addr)
   /* MAC Configuration Control Register should be written last */
   /* Let set_multicast_list set this. */
   /* mace_write(ioaddr, MACE_MACCC, MACE_MACCC_ENXMT | MACE_MACCC_ENRCV); */
+  mace_write(ioaddr, MACE_MACCC, 0x00);
 } /* mace_init */
 
 
@@ -714,13 +847,13 @@ static void nmclan_config(dev_link_t *link)
 
     sig[0] = mace_read(ioaddr, MACE_CHIPIDL);
     sig[1] = mace_read(ioaddr, MACE_CHIPIDH);
-    if ((sig[0] == 0x40) && ((sig[1] == 0x09) || (sig[1] == 0x19))) {
+    if ((sig[0] == 0x40) && ((sig[1] & 0x0F) == 0x09)) {
 #ifdef PCMCIA_DEBUG
       if (pc_debug)
 	printk("nmclan_cs configured: mace id=%x %x\n", sig[0], sig[1]);
 #endif
     } else {
-      printk("nmclan_cs: mace id not found: %x %x should be 0x40 (0x09 or 0x19)\n", sig[0], sig[1]);
+      printk("nmclan_cs: mace id not found: %x %x should be 0x40 0x?9\n", sig[0], sig[1]);
       link->state &= ~DEV_CONFIG_PENDING;
       return;
     }
@@ -743,11 +876,15 @@ static void nmclan_config(dev_link_t *link)
 #if 0
   /* Determine which port we are using if auto is selected */
   if (if_port==4) {
-    if (mace_read(ioaddr, MACE_PHYCC) & /* LNKFL */ 0x80)
+    mace_write(ioaddr, MACE_MACCC, MACE_MACCC_ENXMT | MACE_MACCC_ENRCV);
+    if (nmclan_debug > 2)
+      printk("%s: mace_phycc 0x%X.\n", dev->name, mace_read(ioaddr, MACE_PHYCC));
+    if (mace_read(ioaddr, MACE_PHYCC) & MACE_PHYCC_LNKFL)
       /* 10base-T receiver is in link fail, MACE is using AUI port. */
       dev->if_port = 3;
     else
       dev->if_port = 0;
+    mace_write(ioaddr, MACE_MACCC, 0x00);
   }
   /* Unfortunately, this doesn't seem to work.  LNKFL is always set.
      LNKFL is supposed to be opposite the green LED on the edge of the card.
@@ -755,8 +892,6 @@ static void nmclan_config(dev_link_t *link)
      It does work if check in _start_xmit(), but that's not a good place
      to printk. */
 #endif
-  if (nmclan_debug > 3)
-    printk("%s: mace_phycc 0x%X.\n", dev->name, mace_read(ioaddr, MACE_PHYCC));
 
   printk(version);
   printk("%s: nmclan, io %#3lx, irq %d, %s port,", dev->name,
@@ -861,8 +996,9 @@ static int nmclan_event(event_t event, int priority,
 	link->state &= ~DEV_SUSPEND;
 	CardServices(RequestConfiguration, link->handle, &link->conf);
 	if (link->open) {
+	  dev->tbusy = 0;
+	  dev->start = 1;
 	  nmclan_reset(dev);
-	  dev->tbusy = 0; dev->start = 1;
 	}
       }
       break;
@@ -878,18 +1014,50 @@ static int nmclan_event(event_t event, int priority,
 
 /* ----------------------------------------------------------------------------
 nmclan_reset
-	Reset and restore all of the nmclan registers.
+	Reset and restore all of the Xilinx and MACE registers.
 ---------------------------------------------------------------------------- */
 static void nmclan_reset(struct device *dev)
 {
+  conf_reg_t reg;
+  u_long OrigCorValue; 
+
+
+  /* Save original COR value */
+  reg.Action = CS_READ;
+  reg.Offset = CISREG_COR;
+  reg.Value = 0;
+  CardServices(AccessConfigurationRegister, reg);
+  OrigCorValue = reg.Value;
+
+  /* Reset Xilinx */
+  reg.Action = CS_WRITE;
+  reg.Offset = CISREG_COR;
+  reg.Value = COR_SOFT_RESET;
+  CardServices(AccessConfigurationRegister, reg);
+
+  /* Restore original COR configuration index */
+  reg.Value = COR_LEVEL_REQ | (OrigCorValue & COR_CONFIG_MASK);
+  CardServices(AccessConfigurationRegister, reg);
+
+  /* Xilinx is now completely reset along with the MACE chip. */
+
+  ((mace_private *)dev->priv)->tx_free_frames=AM2150_MAX_TX_FRAMES;
+
+
+  /* Reinitialize the MACE chip for operation. */
   mace_init(dev->base_addr, dev->dev_addr);
   mace_write(dev->base_addr, MACE_IMR, MACE_IMR_DEFAULT);
+
+  /* Restore the multicast list and enable TX and RX. */
+  restore_multicast_list(dev);
 } /* nmclan_reset */
 
 
 /* ----------------------------------------------------------------------------
 mace_config
-	[Someone tell me what this is supposed to do?]
+	[Someone tell me what this is supposed to do?  Is if_port a defined
+	standard?  If so, there should be defines to indicate 0=10Base-T,
+	1=?, 2=?, 3=10Base-2, etc. including limited automatic detection.]
 ---------------------------------------------------------------------------- */
 static int mace_config(struct device *dev, struct ifmap *map)
 {
@@ -903,7 +1071,7 @@ static int mace_config(struct device *dev, struct ifmap *map)
 
 /* ----------------------------------------------------------------------------
 mace_open
-	Open device driver.  set_multicast_list _must_ be called after this.
+	Open device driver.
 ---------------------------------------------------------------------------- */
 static int mace_open(struct device *dev)
 {
@@ -983,20 +1151,33 @@ mace_start_xmit
 ---------------------------------------------------------------------------- */
 static int mace_start_xmit(struct sk_buff *skb, struct device *dev)
 {
+  mace_private *lp = (mace_private *)dev->priv;
   int ioaddr = dev->base_addr;
 
 
-  /* Transmitter timeout, serious problems. */
+#if TIMEOUT_TX
+  /* Transmitter timeout. */
   if (dev->tbusy) {
     int tickssofar = jiffies - dev->trans_start;
 
     if (tickssofar < 40)
       return 1;
-    printk("%s: transmit timed out\n", dev->name);
-    dev->trans_start = jiffies;
+    printk("%s: transmit timed out -- ", dev->name);
+#if RESET_ON_TIMEOUT
+    printk("resetting card\n");
     nmclan_reset(dev);
     dev->tbusy = 0;
+    dev->trans_start = jiffies;
+#else /* #if RESET_ON_TIMEOUT */
+    printk("NOT resetting card\n");
+    dev->trans_start = jiffies;
+    return 1;
+#endif /* #if RESET_ON_TIMEOUT */
   }
+#else /* #if TIMEOUT_TX */
+  if (dev->tbusy)
+    return 1;
+#endif /* #if TIMEOUT_TX */
 
   if (skb == NULL) {
     dev_tint(dev);
@@ -1006,19 +1187,23 @@ static int mace_start_xmit(struct sk_buff *skb, struct device *dev)
   if (skb->len <= 0)
     return 0;
  
+#ifdef NMCLAN_DEBUG
   if (nmclan_debug > 4) {
     printk("%s: mace_start_xmit(length = %ld) called.\n",
       dev->name, skb->len);
   }
- 
+#endif /* #ifdef NMCLAN_DEBUG */
+
   /* Avoid timer-based retransmission conflicts. */
-  if (set_bit(0, (void*)&dev->tbusy) != 0)
+  if (set_bit(0, (void*)&dev->tbusy) != 0) {
     printk("%s: transmitter access conflict.\n", dev->name);
-  else {
+  } else {
+    lp->tx_free_frames--;
+
     /* WARNING: Do not write any more than what is reported! */
-    /* Put out the word header... */
+    /* Put out the word header [must be an outw()] . . . */
     outw(skb->len, ioaddr + AM2150_XMT);
-    /* ... and the packet */
+    /* . . . and the packet [may be any combination of outw() and outb()] */
     outsw(ioaddr + AM2150_XMT, skb->data, skb->len >> 1);
     if (skb->len & 1) {
       /* Odd byte transfer */
@@ -1026,8 +1211,12 @@ static int mace_start_xmit(struct sk_buff *skb, struct device *dev)
     }
 
     dev->trans_start = jiffies;
-    /* dev->tbusy = 0; We are single request right now.
-		       mace_interrupt() will clear it for us. */
+
+    if (lp->tx_free_frames > 0) {
+#if MULTI_TX
+      dev->tbusy = 0;
+#endif /* #if MULTI_TX */
+    }
   }
 
   dev_kfree_skb(skb, FREE_WRITE);
@@ -1045,7 +1234,7 @@ static void mace_interrupt(int irq, struct pt_regs *regs)
   struct device *dev = (struct device *)(irq2dev_map[irq]);
   mace_private *lp = (mace_private *)dev->priv;
   int ioaddr, status;
-  int IntrCnt = 3;
+  int IntrCnt = MACE_MAX_IR_ITERATIONS;
 
 
   if (dev == NULL) {
@@ -1059,83 +1248,122 @@ static void mace_interrupt(int irq, struct pt_regs *regs)
   }
   dev->interrupt = 1;
 
-  ioaddr = dev->base_addr;
-
   if (dev->start == 0) {
     printk("%s: interrupt from dead card\n", dev->name);
     goto exception;
   }
 
-  /* WARNING: MACE_IR is a READ/CLEAR port! */
-  while (
-    (IntrCnt--) &&
-    ((status = inb(ioaddr + AM2150_MACE_BASE + MACE_IR)) &
-      (MACE_IR_XMTINT | MACE_IR_RCVINT))
-  ) {
+
+  ioaddr = dev->base_addr;
+
+  do {
+    /* WARNING: MACE_IR is a READ/CLEAR port! */
+    status = inb(ioaddr + AM2150_MACE_BASE + MACE_IR);
+ 
     if (nmclan_debug > 4) {
       printk("mace_interrupt: irq 0x%X status 0x%X.\n", irq, status);
     }
 
+
     if (status & MACE_IR_RCVINT) {
-      mace_rx(dev, 5);
+      mace_rx(dev, MACE_MAX_RX_ITERATIONS);
     }
 
+
     if (status & MACE_IR_XMTINT) {
       unsigned char fifofc;
+      unsigned char xmtrc;
       unsigned char xmtfs;
 
       fifofc = inb(ioaddr + AM2150_MACE_BASE + MACE_FIFOFC);
       if ((fifofc & MACE_FIFOFC_XMTFC)==0) {
-	lp->stats.tx_errors++;
+	lp->linux_stats.tx_errors++;
 	outb(0xFF, ioaddr + AM2150_XMT_SKIP);
       }
 
-      /* Transmit Retry Count (XMTRC, reg 4), if needed,
-	 must be read here. */
+      /* Transmit Retry Count (XMTRC, reg 4) */
+      xmtrc = inb(ioaddr + AM2150_MACE_BASE + MACE_XMTRC);
+      if (xmtrc & MACE_XMTRC_EXDEF) lp->mace_stats.exdef++;
+      lp->mace_stats.xmtrc += (xmtrc & MACE_XMTRC_XMTRC);
 
       if (
         (xmtfs = inb(ioaddr + AM2150_MACE_BASE + MACE_XMTFS)) &
         MACE_XMTFS_XMTSV /* Transmit Status Valid */
       ) {
-	if (xmtfs & MACE_XMTFS_UFLO) lp->stats.tx_fifo_errors++;
-	  /* Underflow.  Indicates that the Transmit FIFO emptied before
-	     the end of frame was reached. */
-	if (xmtfs & MACE_XMTFS_LCOL) lp->stats.collisions++;
-	  /* Late Collision */
-	/* if (xmtfs & MACE_XMTFS_MORE) More; */
-	  /* MORE than one retry was needed */
-	/* if (xmtfs & MACE_XMTFS_ONE) One; */
-	  /* Exactly ONE retry occurred */
-	/* if (xmtfs & MACE_XMTFS_DEFER) Defer; */
-	  /* Transmission was defered */
-	if (xmtfs & MACE_XMTFS_LCAR) lp->stats.tx_carrier_errors++;
-	  /* Loss of carrier */
-	if (xmtfs & MACE_XMTFS_RTRY) lp->stats.tx_aborted_errors++;
-	  /* Retry error: transmit aborted after 16 attempts */
-      }
+	lp->mace_stats.xmtsv++;
 
-      lp->stats.tx_packets++;
+	if (xmtfs & ~MACE_XMTFS_XMTSV) {
+	  if (xmtfs & MACE_XMTFS_UFLO) {
+	    /* Underflow.  Indicates that the Transmit FIFO emptied before
+	       the end of frame was reached. */
+	    lp->mace_stats.uflo++;
+	  }
+	  if (xmtfs & MACE_XMTFS_LCOL) {
+	    /* Late Collision */
+	    lp->mace_stats.lcol++;
+	  }
+	  if (xmtfs & MACE_XMTFS_MORE) {
+	    /* MORE than one retry was needed */
+	    lp->mace_stats.more++;
+	  }
+	  if (xmtfs & MACE_XMTFS_ONE) {
+	    /* Exactly ONE retry occurred */
+	    lp->mace_stats.one++;
+	  }
+	  if (xmtfs & MACE_XMTFS_DEFER) {
+	    /* Transmission was defered */
+	    lp->mace_stats.defer++;
+	  }
+	  if (xmtfs & MACE_XMTFS_LCAR) {
+	    /* Loss of carrier */
+	    lp->mace_stats.lcar++;
+	  }
+	  if (xmtfs & MACE_XMTFS_RTRY) {
+	    /* Retry error: transmit aborted after 16 attempts */
+	    lp->mace_stats.rtry++;
+	  }
+        } /* if (xmtfs & ~MACE_XMTFS_XMTSV) */
+
+      } /* if (xmtfs & MACE_XMTFS_XMTSV) */
+
+      lp->linux_stats.tx_packets++;
+      lp->tx_free_frames++;
       dev->tbusy = 0;
       mark_bh(NET_BH);
-    }
+    } /* if (status & MACE_IR_XMTINT) */
 
-    if (status & 0x7A) {
-      /* if (status & MACE_IR_JAB) Jabber Error; */
-        /* Jabber Error--Excessive transmit duration (20-150ms) */
-      /* if (status & MACE_IR_BABL) Babble Error; */
-        /* Babble Error-->1518 bytes transmitted */
-      if (status & MACE_IR_CERR) lp->stats.tx_heartbeat_errors++;
+
+    if (status & ~MACE_IMR_DEFAULT & ~MACE_IR_RCVINT & ~MACE_IR_XMTINT) {
+      if (status & MACE_IR_JAB) {
+        /* Jabber Error.  Excessive transmit duration (20-150ms). */
+        lp->mace_stats.jab++;
+      }
+      if (status & MACE_IR_BABL) {
+        /* Babble Error.  >1518 bytes transmitted. */
+        lp->mace_stats.babl++;
+      }
+      if (status & MACE_IR_CERR) {
 	/* Collision Error.  CERR indicates the absence of the
 	   Signal Quality Error Test message after a packet
 	   transmission. */
-      /* if (status & MACE_IR_RCVCCO) Receive Collision Count Overflow; */
-        /* Receive collision count overflow */
-      /* if (status & MACE_IR_RNTPCO) Runt Packet Count Overflow; */
+        lp->mace_stats.cerr++;
+      }
+      if (status & MACE_IR_RCVCCO) {
+        /* Receive Collision Count Overflow; */
+        lp->mace_stats.rcvcco++;
+      }
+      if (status & MACE_IR_RNTPCO) {
         /* Runt Packet Count Overflow */
-      if (status & MACE_IR_MPCO) lp->stats.rx_missed_errors = 255;
+        lp->mace_stats.rntpco++;
+      }
+      if (status & MACE_IR_MPCO) {
         /* Missed Packet Count Overflow */
-    }
-  }
+        lp->mace_stats.mpco++;
+      }
+    } /* if (status & ~MACE_IMR_DEFAULT & ~MACE_IR_RCVINT & ~MACE_IR_XMTINT) */
+
+  } while ((status & ~MACE_IMR_DEFAULT) && (--IntrCnt));
+
 
 exception:
   dev->interrupt = 0;
@@ -1157,7 +1385,7 @@ static int mace_rx(struct device *dev, unsigned char RxCnt)
 
   while (
     ((rx_framecnt = inb(ioaddr + AM2150_RCV_FRAME_COUNT)) > 0) &&
-    (rx_framecnt <= 12) &&
+    (rx_framecnt <= 12) && /* rx_framecnt==0xFF if card is extracted. */
     (RxCnt--)
   ) {
     rx_status = inw(ioaddr + AM2150_RCV);
@@ -1168,25 +1396,35 @@ static int mace_rx(struct device *dev, unsigned char RxCnt)
     }
 
     if (rx_status & MACE_RCVFS_RCVSTS) { /* Error, update stats. */
-      lp->stats.rx_errors++;
-      if (rx_status & MACE_RCVFS_OFLO) lp->stats.rx_over_errors++;
-      /* if (rx_status & MACE_RCVFS_CLSN) lp->stats.collisions++; */
-      if (rx_status & MACE_RCVFS_FRAM) lp->stats.rx_frame_errors++;
-      if (rx_status & MACE_RCVFS_FCS) lp->stats.rx_crc_errors++;
+      lp->linux_stats.rx_errors++;
+      if (rx_status & MACE_RCVFS_OFLO) {
+        lp->mace_stats.oflo++;
+      }
+      if (rx_status & MACE_RCVFS_CLSN) {
+        lp->mace_stats.clsn++;
+      }
+      if (rx_status & MACE_RCVFS_FRAM) {
+	lp->mace_stats.fram++;
+      }
+      if (rx_status & MACE_RCVFS_FCS) {
+        lp->mace_stats.fcs++;
+      }
     } else {
       short pkt_len = (rx_status & ~MACE_RCVFS_RCVSTS) - 4;
         /* Auto Strip is off, always subtract 4 */
       struct sk_buff *skb;
 
-      lp->stats.rx_length_errors += inb(ioaddr + AM2150_RCV);
-	/* runt packet count */
-      lp->stats.collisions += inb(ioaddr + AM2150_RCV);
-	/* rcv collision count */
+      lp->mace_stats.rfs_rntpc += inb(ioaddr + AM2150_RCV);
+        /* runt packet count */
+      lp->mace_stats.rfs_rcvcc += inb(ioaddr + AM2150_RCV);
+        /* rcv collision count */
 
       skb = alloc_skb(pkt_len, GFP_ATOMIC);
+#ifdef NMCLAN_DEBUG
       if (nmclan_debug > 4)
 	printk("    receiving packet size 0x%X rx_status 0x%X.\n",
 	  pkt_len, rx_status);
+#endif /* #ifdef NMCLAN_DEBUG */
       if (skb != NULL) {
 	skb->len = pkt_len;
 	skb->dev = dev;
@@ -1200,38 +1438,108 @@ static int mace_rx(struct device *dev, unsigned char RxCnt)
 
 	netif_rx(skb);
 
-	lp->stats.rx_packets++;
+	lp->linux_stats.rx_packets++;
 	outb(0xFF, ioaddr + AM2150_RCV_NEXT); /* skip to next frame */
-
-	if (nmclan_debug > 5) {
-	  rx_framecnt = inb(ioaddr + AM2150_RCV_FRAME_COUNT);
-	  printk("    1 framecnt 0x%X.\n", rx_framecnt);
-	}
 	continue;
       } else {
         if (nmclan_debug)
 	  printk("%s: couldn't allocate a sk_buff of size %d.\n",
 	    dev->name, pkt_len);
-	lp->stats.rx_dropped++;
+	lp->linux_stats.rx_dropped++;
       }
     }
     outb(0xFF, ioaddr + AM2150_RCV_NEXT); /* skip to next frame */
-
-    if (nmclan_debug > 5) {
-      rx_framecnt = inb(ioaddr + AM2150_RCV_FRAME_COUNT);
-      printk("    2 framecnt 0x%X.\n", rx_framecnt);
-    }
-  }
-
-  if (nmclan_debug > 5) {
-    rx_framecnt = inb(ioaddr + AM2150_RCV_FRAME_COUNT);
-    printk("    3 framecnt 0x%X.\n", rx_framecnt);
-  }
+  } /* while */
 
   return 0;
 } /* mace_rx */
 
 
+/* ----------------------------------------------------------------------------
+pr_linux_stats
+---------------------------------------------------------------------------- */
+static void pr_linux_stats(struct enet_statistics *pstats)
+{
+  printk("pr_linux_stats\n");
+  printk(" rx_packets=%-7d        tx_packets=%d\n",
+    pstats->rx_packets, pstats->tx_packets);
+  printk(" rx_errors=%-7d         tx_errors=%d\n",
+    pstats->rx_errors, pstats->tx_errors);
+  printk(" rx_dropped=%-7d        tx_dropped=%d\n",
+    pstats->rx_dropped, pstats->tx_dropped);
+  printk(" multicast=%-7d         collisions=%d\n",
+    pstats->multicast, pstats->collisions);
+
+  printk(" rx_length_errors=%-7d  rx_over_errors=%d\n",
+    pstats->rx_length_errors, pstats->rx_over_errors);
+  printk(" rx_crc_errors=%-7d     rx_frame_errors=%d\n",
+    pstats->rx_crc_errors, pstats->rx_frame_errors);
+  printk(" rx_fifo_errors=%-7d    rx_missed_errors=%d\n",
+    pstats->rx_fifo_errors, pstats->rx_missed_errors);
+
+  printk(" tx_aborted_errors=%-7d tx_carrier_errors=%d\n",
+    pstats->tx_aborted_errors, pstats->tx_carrier_errors);
+  printk(" tx_fifo_errors=%-7d    tx_heartbeat_errors=%d\n",
+    pstats->tx_fifo_errors, pstats->tx_heartbeat_errors);
+  printk(" tx_window_errors=%d\n",
+    pstats->tx_window_errors);
+} /* pr_linux_stats */
+
+
+/* ----------------------------------------------------------------------------
+pr_mace_stats
+---------------------------------------------------------------------------- */
+static void pr_mace_stats(mace_statistics *pstats)
+{
+  printk("pr_mace_stats\n");
+
+  printk(" xmtsv=%-7d             uflo=%d\n",
+    pstats->xmtsv, pstats->uflo);
+  printk(" lcol=%-7d              more=%d\n",
+    pstats->lcol, pstats->more);
+  printk(" one=%-7d               defer=%d\n",
+    pstats->one, pstats->defer);
+  printk(" lcar=%-7d              rtry=%d\n",
+    pstats->lcar, pstats->rtry);
+
+  /* MACE_XMTRC */
+  printk(" exdef=%-7d             xmtrc=%d\n",
+    pstats->exdef, pstats->xmtrc);
+
+  /* RFS1--Receive Status (RCVSTS) */
+  printk(" oflo=%-7d              clsn=%d\n",
+    pstats->oflo, pstats->clsn);
+  printk(" fram=%-7d              fcs=%d\n",
+    pstats->fram, pstats->fcs);
+
+  /* RFS2--Runt Packet Count (RNTPC) */
+  /* RFS3--Receive Collision Count (RCVCC) */
+  printk(" rfs_rntpc=%-7d         rfs_rcvcc=%d\n",
+    pstats->rfs_rntpc, pstats->rfs_rcvcc);
+
+  /* MACE_IR */
+  printk(" jab=%-7d               babl=%d\n",
+    pstats->jab, pstats->babl);
+  printk(" cerr=%-7d              rcvcco=%d\n",
+    pstats->cerr, pstats->rcvcco);
+  printk(" rntpco=%-7d            mpco=%d\n",
+    pstats->rntpco, pstats->mpco);
+
+  /* MACE_MPC */
+  printk(" mpc=%d\n",
+    pstats->mpc);
+
+  /* MACE_RNTPC */
+  printk(" rntpc=%d\n",
+    pstats->rntpc);
+
+  /* MACE_RCVCC */
+  printk(" rcvcc=%d\n",
+    pstats->rcvcc);
+
+} /* pr_mace_stats */
+
+
 /* ----------------------------------------------------------------------------
 update_stats
 	Update statistics.  We change to register window 1, so this
@@ -1239,16 +1547,61 @@ update_stats
 	expected to be a rare operation, and it's simpler for the rest
 	of the driver to assume that window 0 is always valid rather
 	than use a special window-state variable.
+
+	oflo & uflo should _never_ occur since it would mean the Xilinx
+	was not able to transfer data between the MACE FIFO and the
+	card's SRAM fast enough.  If this happens, something is
+	seriously wrong with the hardware.
 ---------------------------------------------------------------------------- */
 static void update_stats(int ioaddr, struct device *dev)
 {
   mace_private *lp = (mace_private *)dev->priv;
     
 
-  if (nmclan_debug > 5)
-    printk("%s: updating the statistics.\n", dev->name);
+  lp->mace_stats.rcvcc += mace_read(ioaddr, MACE_RCVCC);
+  lp->mace_stats.rntpc += mace_read(ioaddr, MACE_RNTPC);
+  lp->mace_stats.mpc += mace_read(ioaddr, MACE_MPC);
+  /* At this point, mace_stats is fully updated for this call.
+     We may now update the linux_stats. */
 
-  lp->stats.rx_missed_errors += mace_read(ioaddr, MACE_MPC);
+
+  /* The MACE has no equivalent for linux_stats field which are commented
+     out. */
+
+#if 0
+  /* These must be tracked in the main body of the driver. */
+  lp->linux_stats.rx_packets;
+  lp->linux_stats.tx_packets;
+  lp->linux_stats.rx_errors;
+  lp->linux_stats.tx_errors;
+  lp->linux_stats.rx_dropped;
+  lp->linux_stats.tx_dropped;
+#endif
+  /* lp->linux_stats.multicast; */
+  lp->linux_stats.collisions = 
+    lp->mace_stats.rcvcco * 256 + lp->mace_stats.rcvcc;
+    /* Collision: The MACE may retry sending a packet 15 times
+       before giving up.  The retry count is in XMTRC.
+       Does each retry constitute a collision?
+       If so, why doesn't the RCVCC record these collisions? */
+
+  /* detailed rx_errors: */
+  lp->linux_stats.rx_length_errors = 
+    lp->mace_stats.rntpco * 256 + lp->mace_stats.rntpc;
+  /* lp->linux_stats.rx_over_errors */
+  lp->linux_stats.rx_crc_errors = lp->mace_stats.fcs;
+  lp->linux_stats.rx_frame_errors = lp->mace_stats.fram;
+  lp->linux_stats.rx_fifo_errors = lp->mace_stats.oflo;
+  lp->linux_stats.rx_missed_errors = 
+    lp->mace_stats.mpco * 256 + lp->mace_stats.mpc;
+
+  /* detailed tx_errors */
+  lp->linux_stats.tx_aborted_errors = lp->mace_stats.rtry;
+  lp->linux_stats.tx_carrier_errors = lp->mace_stats.lcar;
+    /* LCAR usually results from bad cabling. */
+  lp->linux_stats.tx_fifo_errors = lp->mace_stats.uflo;
+  lp->linux_stats.tx_heartbeat_errors = lp->mace_stats.cerr;
+  /* lp->linux_stats.tx_window_errors; */
 
   return;
 } /* update_stats */
@@ -1261,14 +1614,17 @@ mace_get_stats
 static struct enet_statistics *mace_get_stats(struct device *dev)
 {
   mace_private *lp = (mace_private *)dev->priv;
-  unsigned long flags;
 
 
-  save_flags(flags);
-  cli();
   update_stats(dev->base_addr, dev);
-  restore_flags(flags);
-  return &lp->stats;
+
+  if (nmclan_debug > 2 /* TBD: 5 */) {
+    printk("%s: updating the statistics.\n", dev->name);
+    pr_linux_stats(&lp->linux_stats);
+    pr_mace_stats(&lp->mace_stats);
+  }
+
+  return &lp->linux_stats;
 } /* enet_statistics */
 
 
@@ -1344,41 +1700,36 @@ static void BuildLAF(int *ladrf, int *adr)
 
 
 /* ----------------------------------------------------------------------------
-set_multicast_list
-	Set or clear the multicast filter for this adaptor.
+restore_multicast_list
+	Restores the multicast filter for MACE chip to the last
+	set_multicast_list() call.
 
 Input
-	num_addrs == -1	Promiscuous mode, receive all packets
-	num_addrs == 0	Normal mode, clear multicast list
-	num_addrs > 0	Multicast mode, receive normal and MC packets, and do
-			best-effort filtering.
+	multicast_num_addrs
+	multicast_ladrf[]
 ---------------------------------------------------------------------------- */
-static void set_multicast_list(struct device *dev, int num_addrs, void *addrs)
+static void restore_multicast_list(struct device *dev)
 {
+  mace_private *lp = (mace_private *)dev->priv;
+  int num_addrs = lp->multicast_num_addrs;
+  int *ladrf = lp->multicast_ladrf;
   short ioaddr = dev->base_addr;
-  int adr[6] = {0}; /* Ethernet address */
-  int ladrf[8] = {0}; /* Logical address filter */
   int i;
 
 
-  if (nmclan_debug > 1)
-    printk("%s: setting Rx mode to %d addresses.\n", dev->name, num_addrs);
+  if (nmclan_debug > 3)
+    printk("%s: restoring Rx mode to %d addresses.\n", dev->name, num_addrs);
 
   if (num_addrs > 0) {
 
-    printk("Attempt to set multicast list detected.\n");
+    printk("Attempt to restore multicast list detected.\n");
 
-    /* Calculate multicast logical address filter */
-    for (i = 0; i < num_addrs; i++) {
-      /* TBD: Set adr[] to the i'th address */
-      BuildLAF(ladrf, adr);
-    }
     mace_write(ioaddr, MACE_IAC, MACE_IAC_ADDRCHG | MACE_IAC_LOGADDR);
     /* Poll ADDRCHG bit */
     while (mace_read(ioaddr, MACE_IAC) & MACE_IAC_ADDRCHG)
       ;
     /* Set LADRF register */
-    for (i = 0; i < ETHER_ADDR_LEN; i++)
+    for (i = 0; i < MACE_LADRF_LEN; i++)
       mace_write(ioaddr, MACE_LADRF, ladrf[i]);
 
     mace_write(ioaddr, MACE_UTR, MACE_UTR_RCVFCSE | MACE_UTR_LOOP_EXTERNAL);
@@ -1402,6 +1753,57 @@ static void set_multicast_list(struct device *dev, int num_addrs, void *addrs)
 } /* set_multicast_list */
 
 
+/* ----------------------------------------------------------------------------
+set_multicast_list
+	Set or clear the multicast filter for this adaptor.
+
+Input
+	num_addrs == -1	Promiscuous mode, receive all packets
+	num_addrs == 0	Normal mode, clear multicast list
+	num_addrs > 0	Multicast mode, receive normal and MC packets, and do
+			best-effort filtering.
+Output
+	multicast_num_addrs
+	multicast_ladrf[]
+---------------------------------------------------------------------------- */
+static void set_multicast_list(struct device *dev, int num_addrs, void *addrs)
+{
+  mace_private *lp = (mace_private *)dev->priv;
+  int adr[ETHER_ADDR_LEN] = {0}; /* Ethernet address */
+  int i;
+
+
+  if (nmclan_debug > 1)
+    printk("%s: setting Rx mode to %d addresses.\n", dev->name, num_addrs);
+
+
+  /* Set multicast_num_addrs. */
+  lp->multicast_num_addrs = num_addrs;
+
+
+  /* Set multicast_ladrf. */
+  if (num_addrs > 0) {
+    printk("Attempt to set multicast list detected.\n");
+
+    /* Calculate multicast logical address filter */
+    memset(lp->multicast_ladrf, 0, MACE_LADRF_LEN);
+    for (i = 0; i < num_addrs; i++) {
+      int j;
+      char *pa=(char *)addrs;
+
+      /* Set adr[] to the i'th address */
+      for (j = 0; j < ETHER_ADDR_LEN; j++)
+	adr[j] = pa[ETHER_ADDR_LEN*i+j];
+      BuildLAF(lp->multicast_ladrf, adr);
+    }
+  }
+
+
+  restore_multicast_list(dev);
+
+} /* set_multicast_list */
+
+
 /* ------------------------------------------------------------------------- */
 
 #ifdef MODULE
