@@ -28,6 +28,7 @@
 #include <asm/io.h>
 #include <asm/system.h>
 #include <asm/segment.h>
+#include <asm/string.h>
 #include <stdarg.h>
 
 #include "version.h"
@@ -53,7 +54,7 @@ static int major_dev = 0;
 #ifdef PCMCIA_DEBUG
 static int pc_debug = PCMCIA_DEBUG;
 static char *version =
-"pcmem_cs.c 1.41 1995/05/15 01:43:40 (David Hinds)\n";
+"pcmem_cs.c 1.42 1995/05/24 06:24:26 (David Hinds)\n";
 #endif
 
 /*====================================================================*/
@@ -364,7 +365,7 @@ static u_long get_size(dev_link_t *link, pcmem_dev_t *dev)
 
     pcmem_config() is scheduled to run after a CARD_INSERTION event
     is received, to configure the PCMCIA socket, and to make the
-    ethernet device available to the system.
+    memory device available to the system.
     
 ======================================================================*/
 
@@ -388,7 +389,7 @@ static void pcmem_config(dev_link_t *link)
 
     do {
 	/* Allocate a 4K memory window */
-	req.Attributes = WIN_DATA_WIDTH_8;
+	req.Attributes = WIN_DATA_WIDTH_16;
 	req.Base = NULL; req.Size = WINDOW_SIZE;
 	req.AccessSpeed = 0;
 	link->win = (window_handle_t)link->handle;
@@ -729,7 +730,7 @@ static int pcmem_ioctl(struct inode *inode, struct file *file,
 
 static void do_pcmem_request(void)
 {
-    int addr, len, from, nb, i, ret;
+    int addr, len, from, nb, ret;
     char *buf;
     dev_link_t *link;
     pcmem_dev_t *dev;
@@ -776,8 +777,7 @@ static void do_pcmem_request(void)
 		ret = CardServices(MapMemPage, link->win, &mem);
 		if (ret != CS_SUCCESS) break;
 		nb = (from+len > WINDOW_SIZE) ? WINDOW_SIZE-from : len;
-		for (i = 0; i < nb; i++, buf++)
-		    *buf = dev->Base[from+i];
+		memcpy(buf, &dev->Base[from], nb);
 		mem.CardOffset += WINDOW_SIZE;
 	    }
 	    break;
@@ -787,8 +787,7 @@ static void do_pcmem_request(void)
 		ret = CardServices(MapMemPage, link->win, &mem);
 		if (ret != CS_SUCCESS) break;
 		nb = (from+len > WINDOW_SIZE) ? WINDOW_SIZE-from : len;
-		for (i = 0; i < nb; i++, buf++)
-		    dev->Base[from+i] = *buf;
+		memcpy(&dev->Base[from], buf, nb);
 		mem.CardOffset += WINDOW_SIZE;
 	    }
 	    break;
